# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from xypattern import Pattern
from xypattern.pattern import SmoothBrucknerBackground

from .util import Signal
from .util.HelperModule import FileNameIterator, get_base_name

logger = logging.getLogger(__name__)


class PatternModel(object):
    """
    Main Pattern handling class. Supporting:
        - setting background pattern
        - setting automatic background subtraction
        - file browsing

    all changes to the internal data throw a pattern_changed signal.
    """

    def __init__(self):
        super(PatternModel, self).__init__()
        self.pattern = Pattern()
        self.pattern_filename = ""

        self.unit = ""
        self.file_iteration_mode = "number"
        self.file_name_iterator = FileNameIterator()

        self._background_pattern = None

        self.pattern_changed = Signal()

    def set_pattern(self, x, y, filename="", unit=""):
        """
        set the current data pattern.
        :param x: x-values
        :param y: y-values
        :param filename: name for the pattern, defaults to ''
        :param unit: unit for the x values
        """
        self.pattern_filename = filename
        self.pattern.data = (x, y)
        self.pattern.name = get_base_name(filename)
        self.unit = unit
        self.pattern_changed.emit()

    def load_pattern(self, filename):
        """
        Loads a pattern from a tabular pattern file (2 column txt file)
        :param filename: filename of the data file
        """
        logger.info("Load pattern: {0}".format(filename))
        self.pattern_filename = filename

        skiprows = 0
        if filename.endswith(".chi"):
            skiprows = 4
        self.pattern.load(filename, skiprows)
        self.file_name_iterator.update_filename(filename)
        self.pattern_changed.emit()

    def save_pattern(
        self, filename: str, header: str = "", subtract_background: bool = False
    ):
        """
        Saves the current data pattern.
        :param filename: where to save
        :param header: you can specify any specific header
        :param subtract_background: whether the background set will be used for saving or not
        """
        self.pattern.save(filename, header, subtract_background, self.unit)

    def save_auto_background_as_pattern(self, filename, header=None):
        """
        Saves the current automatic extracted background data pattern to file
        :param filename: where to save
        :param header: you can specify any specific header
        """
        self.pattern.auto_background_pattern.save(filename, header, unit=self.unit)

    def get_pattern(self):
        return self.pattern

    def load_next_file(self, step=1):
        """
        Loads the next file from a sequel of filenames (e.g. *_001.xy --> *_002.xy)
        It assumes that the file numbers are at the end of the filename
        """
        next_file_name = self.file_name_iterator.get_next_filename(
            mode=self.file_iteration_mode, step=step
        )
        if next_file_name is not None:
            self.load_pattern(next_file_name)
            return True
        return False

    def load_previous_file(self, step=1):
        """
        Loads the previous file from a sequel of filenames (e.g. *_002.xy --> *_001.xy)
        It assumes that the file numbers are at the end of the filename
        """
        next_file_name = self.file_name_iterator.get_previous_filename(
            mode=self.file_iteration_mode, step=step
        )
        if next_file_name is not None:
            self.load_pattern(next_file_name)
            return True
        return False

    def set_file_iteration_mode(self, mode):
        if mode == "number":
            self.file_iteration_mode = "number"
            self.file_name_iterator.create_timed_file_list = False
        elif mode == "time":
            self.file_iteration_mode = "time"
            self.file_name_iterator.create_timed_file_list = True
            self.file_name_iterator.update_filename(self.pattern_filename)

    @property
    def background_pattern(self):
        return self._background_pattern

    @background_pattern.setter
    def background_pattern(self, pattern: Pattern | None):
        if pattern is not None:
            self.pattern.background_pattern = pattern
        else:
            self.pattern.background_pattern = None
        self._background_pattern = pattern
        self.pattern_changed.emit()

    def set_auto_background_subtraction(self, parameters, roi=None):
        """
        Enables auto background extraction and removal from the data pattern
        :param parameters: array of parameters with [window_width, iterations, polynomial_order]
        :param roi: array of size two with [x_min, x_max] specifying the range for the background subtraction
        will be performed
        """
        if roi is not None:
            x, _ = self.pattern.original_data
            roi = list(roi)

            # make sure the roi is within the data range
            if roi[0] > roi[1]:
                roi[0], roi[1] = roi[1], roi[0]
            x_step = x[1] - x[0]
            roi[0] = roi[0] if roi[0] > x[0] else x[0] - x_step / 2
            roi[0] = roi[0] if roi[0] < x[-1] - 1.5 * x_step else x[-1] - 1.5 * x_step
            roi[1] = roi[1] if roi[1] < x[-1] else x[-1] + x_step / 2
            roi[1] = roi[1] if roi[1] > x[0] + 1.5 * x_step else x[0] + 1.5 * x_step

        self.pattern.auto_bkg_roi = roi
        self.pattern.auto_bkg = SmoothBrucknerBackground(*parameters)
        self.pattern_changed.emit()

    def unset_auto_background_subtraction(self):
        """
        Disables auto background extraction and removal.
        """
        self.pattern.auto_bkg = None
        self.pattern_changed.emit()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from functools import partial

import numpy as np
from qtpy import QtWidgets, QtCore, QtGui

from ....widgets.UtilityWidgets import save_file_dialog
from ....widgets.integration.JcpdsEditorWidget import JcpdsEditorWidget

# imports for type hinting in PyCharm -- DO NOT DELETE
from ....model.util.jcpds import jcpds, jcpds_reflection
from ....model.DioptasModel import DioptasModel
from ....widgets.integration import IntegrationWidget


class JcpdsEditorController(QtCore.QObject):
    """
    JcpdsEditorController handles all the signals and changes associated with Jcpds editor widget
    """
    canceled_editor = QtCore.Signal(jcpds)

    # lattice_param_changed = QtCore.Signal()
    # eos_param_changed = QtCore.Signal()
    #
    # reflection_line_edited = QtCore.Signal()
    # reflection_line_removed = QtCore.Signal(int)
    # reflection_line_cleared = QtCore.Signal()

    def __init__(self, integration_widget, dioptas_model=None, jcpds_phase=None):
        """
        :param integration_widget: Reference to an IntegrationWidget
        :param dioptas_model: Reference to DioptasModel object
        :param jcpds_phase: Reference to JcpdsPhase object

        :type integration_widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        :type jcpds_phase: jcpds
        """
        super(JcpdsEditorController, self).__init__()
        self.integration_widget = integration_widget
        self.jcpds_widget = JcpdsEditorWidget(integration_widget)
        self.phase_widget = self.integration_widget.phase_widget
        self.model = dioptas_model
        self.phase_model = self.model.phase_model
        self.create_connections()

        self.previous_header_item_index_sorted = None
        self.active = False
        self.phase_ind = -1

        if jcpds_phase is not None:
            self.show_phase(jcpds_phase)

    def show_phase(self, jcpds_phase=None, wavelength=None):
        self.start_jcpds_phase = deepcopy(jcpds_phase)
        self.jcpds_phase = jcpds_phase
        self.phase_ind = self.phase_model.phases.index(jcpds_phase)
        if wavelength is None:
            if self.model.calibration_model is not None:
                wavelength = self.model.calibration_model.wavelength * 1e10
        self.jcpds_widget.show_jcpds(jcpds_phase, wavelength)

    def create_connections(self):
        # Phase Widget Signals
        self.phase_widget.edit_btn.clicked.connect(self.edit_btn_callback)
        self.phase_widget.phase_tw.currentCellChanged.connect(self.phase_selection_changed)

        # Phase Model signals
        self.phase_model.phase_changed.connect(self.phase_changed)
        self.phase_model.reflection_added.connect(self.update_reflection_table)
        self.phase_model.reflection_deleted.connect(self.update_reflection_table)

        # Information fields
        self.jcpds_widget.comments_txt.editingFinished.connect(self.comments_changed)
        self.jcpds_widget.symmetry_cb.currentIndexChanged.connect(self.symmetry_changed)
        #
        # Lattice Parameter fields
        self.jcpds_widget.lattice_a_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                    widget=self.jcpds_widget.lattice_a_sb,
                                                                    param='a0'))
        self.jcpds_widget.lattice_b_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                    widget=self.jcpds_widget.lattice_b_sb,
                                                                    param='b0'))
        self.jcpds_widget.lattice_c_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                    widget=self.jcpds_widget.lattice_c_sb,
                                                                    param='c0'))

        self.jcpds_widget.lattice_ab_sb.valueChanged.connect(self.lattice_ab_changed)
        self.jcpds_widget.lattice_ca_sb.valueChanged.connect(self.lattice_ca_changed)
        self.jcpds_widget.lattice_cb_sb.valueChanged.connect(self.lattice_cb_changed)

        self.jcpds_widget.lattice_alpha_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                        widget=self.jcpds_widget.lattice_alpha_sb,
                                                                        param='alpha0'))
        self.jcpds_widget.lattice_beta_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                       widget=self.jcpds_widget.lattice_beta_sb,
                                                                       param='beta0'))
        self.jcpds_widget.lattice_gamma_sb.valueChanged.connect(partial(self.param_sb_changed,
                                                                        widget=self.jcpds_widget.lattice_gamma_sb,
                                                                        param='gamma0'))

        self.jcpds_widget.lattice_length_step_txt.editingFinished.connect(self.lattice_length_step_changed)
        self.jcpds_widget.lattice_angle_step_txt.editingFinished.connect(self.lattice_angle_step_changed)
        self.jcpds_widget.lattice_ratio_step_txt.editingFinished.connect(self.lattice_ratio_step_changed)

        # Equation of state fields
        self.jcpds_widget.eos_K_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                    widget=self.jcpds_widget.eos_K_txt,
                                                                    param='k0'))
        self.jcpds_widget.eos_Kp_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                     widget=self.jcpds_widget.eos_Kp_txt,
                                                                     param='k0p0'))
        self.jcpds_widget.eos_alphaT_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                         widget=self.jcpds_widget.eos_alphaT_txt,
                                                                         param='alpha_t0'))
        self.jcpds_widget.eos_dalphadT_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                           widget=self.jcpds_widget.eos_dalphadT_txt,
                                                                           param='d_alpha_dt'))
        self.jcpds_widget.eos_dKdT_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                       widget=self.jcpds_widget.eos_dKdT_txt,
                                                                       param='dk0dt'))
        self.jcpds_widget.eos_dKpdT_txt.editingFinished.connect(partial(self.param_txt_changed,
                                                                        widget=self.jcpds_widget.eos_dKpdT_txt,
                                                                        param='dk0pdt'))

        # Reflections Controls
        self.jcpds_widget.reflections_add_btn.clicked.connect(self.reflections_add_btn_click)
        self.jcpds_widget.reflections_delete_btn.clicked.connect(self.reflections_delete_btn_click)
        self.jcpds_widget.reflections_clear_btn.clicked.connect(self.reflections_clear_btn_click)
        self.jcpds_widget.reflection_table_model.reflection_edited.connect(self.reflection_table_changed)

        # Table Widgets events
        self.jcpds_widget.reflection_table_view.keyPressEvent = self.reflection_table_key_pressed
        self.jcpds_widget.reflection_table_view.verticalScrollBar().valueChanged.connect(self.reflection_table_scrolled)
        self.jcpds_widget.reflection_table_view.horizontalHeader().sectionClicked.connect(
            self.horizontal_header_clicked)
        #
        # Button fields
        self.jcpds_widget.reload_file_btn.clicked.connect(self.reload_file_btn_clicked)
        self.jcpds_widget.save_as_btn.clicked.connect(self.save_as_btn_clicked)
        #
        # # Closing and opening
        self.jcpds_widget.closeEvent = self.view_closed

    def edit_btn_callback(self):
        selected_row = self.phase_widget.get_selected_phase_row()
        if selected_row >= 0:
            self.show_phase(self.model.phase_model.phases[selected_row])
            self.show_view()

    def phase_selection_changed(self, row, *_):
        if self.active:
            self.show_phase(self.model.phase_model.phases[row])

    def phase_changed(self, ind):
        if self.active and self.phase_ind == ind:
            self.jcpds_widget.show_jcpds(self.phase_model.phases[ind],
                                         wavelength=self.model.calibration_model.wavelength * 1e10)

    def update_filename(self):
        self.jcpds_widget.filename_txt.setText(self.jcpds_phase.filename)

    def comments_changed(self):
        self.jcpds_phase.params['comments'][0] = str(self.jcpds_widget.comments_txt.text())

    def symmetry_changed(self):
        self.phase_model.set_param(self.phase_ind, 'symmetry',
                                   str(self.jcpds_widget.symmetry_cb.currentText()).upper())

    def param_sb_changed(self, widget, param):
        self.phase_model.set_param(self.phase_ind, param, widget.value())

    def param_txt_changed(self, widget, param):
        self.phase_model.set_param(self.phase_ind, param, float(widget.text()))

    def lattice_ab_changed(self):
        ab_ratio = float(self.jcpds_widget.lattice_ab_sb.value())
        self.phase_model.set_param(self.phase_ind, 'a0',
                                   self.jcpds_phase.params['b0'] * ab_ratio)

    def lattice_ca_changed(self):
        ca_ratio = float(self.jcpds_widget.lattice_ca_sb.value())
        self.phase_model.set_param(self.phase_ind, 'c0',
                                   self.jcpds_phase.params['a0'] * ca_ratio)

    def lattice_cb_changed(self):
        cb_ratio = float(self.jcpds_widget.lattice_cb_sb.value())
        self.phase_model.set_param(self.phase_ind, 'c0',
                                   self.jcpds_phase.params['b0'] * cb_ratio)

    def lattice_length_step_changed(self):
        value = float(str(self.jcpds_widget.lattice_length_step_txt.text()))
        self.jcpds_widget.lattice_a_sb.setSingleStep(value)
        self.jcpds_widget.lattice_b_sb.setSingleStep(value)
        self.jcpds_widget.lattice_c_sb.setSingleStep(value)

    def lattice_angle_step_changed(self):
        value = float(str(self.jcpds_widget.lattice_angle_step_txt.text()))
        self.jcpds_widget.lattice_alpha_sb.setSingleStep(value)
        self.jcpds_widget.lattice_beta_sb.setSingleStep(value)
        self.jcpds_widget.lattice_gamma_sb.setSingleStep(value)

    def lattice_ratio_step_changed(self):
        value = float(str(self.jcpds_widget.lattice_ratio_step_txt.text()))
        self.jcpds_widget.lattice_ab_sb.setSingleStep(value)
        self.jcpds_widget.lattice_ca_sb.setSingleStep(value)
        self.jcpds_widget.lattice_cb_sb.setSingleStep(value)

    def reflections_delete_btn_click(self):
        rows = self.jcpds_widget.get_selected_reflections()
        self.phase_model.delete_multiple_reflections(self.phase_ind, rows)
        if self.jcpds_widget.reflection_table_model.rowCount() >= min(rows) + 1:
            self.jcpds_widget.reflection_table_view.selectRow(min(rows))
        else:
            self.jcpds_widget.reflection_table_view.selectRow(
                self.jcpds_widget.reflection_table_model.rowCount() - 1)

    def reflections_add_btn_click(self):
        self.phase_model.add_reflection(self.phase_ind)
        self.jcpds_widget.reflection_table_view.selectRow(self.jcpds_widget.reflection_table_model.rowCount() - 1)

    def reflections_clear_btn_click(self):
        self.phase_model.clear_reflections(self.phase_ind)

    def reflection_table_changed(self, row, column, value):
        if value != '':
            value = float(value)
            reflection = self.phase_model.phases[self.phase_ind].reflections[row]
            if column == 0:  # h
                reflection.h = value
            elif column == 1:  # k
                reflection.k = value
            elif column == 2:  # l
                reflection.l = value
            elif column == 3:  # intensity
                reflection.intensity = value
            self.phase_model.update_reflection(self.phase_ind, row, reflection)
            self.jcpds_widget.reflection_table_model.update_reflection_data(
                self.phase_model.phases[self.phase_ind].reflections)

    def update_reflection_table(self, phase_ind, *_):
        if phase_ind != self.phase_ind:
            return
        self.jcpds_widget.reflection_table_model.update_reflection_data(self.phase_model.phases[phase_ind].reflections)

    def reflection_table_key_pressed(self, key_press_event):
        if key_press_event == QtGui.QKeySequence.Copy:
            select = self.jcpds_widget.reflection_table_view.selectionModel()
            if not select.hasSelection():
                return # nothing selected

            lines = []
            if self.jcpds_widget.reflection_table_model.wavelength is None:
                lines.append('\t'.join(self.jcpds_widget.reflection_table_model.header_labels)) # Header
                for item in select.selectedRows():
                    line = '{:.0f}\t{:.0f}\t{:.0f}\t{:.2f}\t{:.4f}\t{:.4f}'.format(
                        *self.jcpds_widget.reflection_table_model.reflection_data[item.row(), :]
                    )
                    lines.append(line)
            else:
                lines.append('\t'.join(self.jcpds_widget.reflection_table_model.header_labels)) # Header
                for item in select.selectedRows():
                    line = '{:.0f}\t{:.0f}\t{:.0f}\t{:.2f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}'.format(
                        *self.jcpds_widget.reflection_table_model.reflection_data[item.row(), :]
                    )
                    lines.append(line)
            QtWidgets.QApplication.clipboard().setText('\n'.join(lines))

        elif key_press_event == QtGui.QKeySequence.SelectAll:
            self.jcpds_widget.reflection_table_view.selectAll()

    def reflection_table_scrolled(self):
        self.jcpds_widget.reflection_table_view.resizeColumnsToContents()

    def horizontal_header_clicked(self, ind):
        if self.previous_header_item_index_sorted == ind:
            reversed_toggle = True
        else:
            reversed_toggle = False

        if ind == 0:
            self.jcpds_phase.sort_reflections_by_h(reversed_toggle)
        elif ind == 1:
            self.jcpds_phase.sort_reflections_by_k(reversed_toggle)
        elif ind == 2:
            self.jcpds_phase.sort_reflections_by_l(reversed_toggle)
        elif ind == 3:
            self.jcpds_phase.sort_reflections_by_intensity(reversed_toggle)
        elif ind == 4 or ind == 6:
            self.jcpds_phase.sort_reflections_by_d(reversed_toggle)
        elif ind == 5 or ind == 7:
            self.jcpds_phase.sort_reflections_by_d(not reversed_toggle)

        self.jcpds_widget.show_jcpds(self.jcpds_phase, wavelength=self.model.calibration_model.wavelength * 1e10)
        self.jcpds_widget.reflection_table_view.resizeColumnsToContents()

        if self.previous_header_item_index_sorted == ind:
            self.previous_header_item_index_sorted = None
        else:
            self.previous_header_item_index_sorted = ind

    def save_as_btn_clicked(self, filename=False):
        if filename is False:
            filename = save_file_dialog(self.jcpds_widget, "Save JCPDS phase.",
                                        self.model.working_directories['phase'],
                                        ('JCPDS Phase (*.jcpds);;Export Table (*.txt)'))

            if filename != '':
                if filename.endswith('.jcpds'):
                    self.phase_model.save_phase_as(self.phase_ind, filename)
                elif filename.endswith('.txt'):
                    self.export_table_data(filename)
            self.show_phase(self.jcpds_phase)

    def export_table_data(self, filename):
        fp = open(filename, 'w', encoding='utf-8')
        for col in range(self.jcpds_widget.reflection_table_model.columnCount()):
            fp.write(self.jcpds_widget.reflection_table_model.header_labels[col] + '\t')
        fp.write('\n')
        for row in range(self.jcpds_widget.reflection_table_model.rowCount()):
            line = ''
            for col in range(self.jcpds_widget.reflection_table_model.columnCount()):
                line = line + self.jcpds_widget.reflection_table_model.index(row, col).data() + '\t'
            line = line + '\n'
            fp.write(line)
        fp.close()

    def reload_file_btn_clicked(self):
        self.phase_model.reload(self.phase_ind)

    def show_view(self):
        self.active = True
        self.jcpds_widget.raise_widget()

    def close_view(self):
        self.active = False
        self.jcpds_widget.close()

    def view_closed(self, _):
        self.close_view()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

CLICKED_COLOR = '#00DD00'

from qtpy import QtWidgets, QtCore

from ..UtilityWidgets import FileInfoWidget
from ..EpicsWidgets import MoveStageWidget
from .BatchWidget import BatchWidget

from .CustomWidgets import MouseCurrentAndClickedWidget, MouseUnitCurrentAndClickedWidget
from .control import IntegrationControlWidget
from .display.ImgWidget import IntegrationImgDisplayWidget
from .display.PatternWidget import IntegrationPatternWidget
from .display.StatusWidget import IntegrationStatusWidget


class IntegrationWidget(QtWidgets.QWidget):
    """
    Defines the main structure of the integration widget, which is separated into four parts.
    Integration Image Widget - displaying the image, mask and/or cake
    Integration Control Widget - Handling all the interaction with overlays, phases etc.
    Integration Pattern Widget - showing the integrated pattern
    Integration Status Widget - showing the current mouse position and current background filename
    """

    def __init__(self, *args, **kwargs):
        super(IntegrationWidget, self).__init__(*args, **kwargs)

        self.setObjectName('integration_widget')

        self.integration_image_widget = IntegrationImgDisplayWidget()
        self.integration_control_widget = IntegrationControlWidget()
        self.integration_pattern_widget = IntegrationPatternWidget()
        self.integration_status_widget = IntegrationStatusWidget()

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.vertical_splitter = QtWidgets.QSplitter(self)
        self.vertical_splitter.setOrientation(QtCore.Qt.Vertical)
        self.vertical_splitter.addWidget(self.integration_control_widget)
        self.vertical_splitter.addWidget(self.integration_pattern_widget)
        self.vertical_splitter.setStretchFactor(0, 1)
        self.vertical_splitter.setStretchFactor(1, 10)

        self.vertical_splitter_left = QtWidgets.QSplitter(self)
        self.vertical_splitter_left.setOrientation(QtCore.Qt.Vertical)
        self.vertical_splitter_left.addWidget(self.integration_image_widget)

        self.horizontal_splitter = QtWidgets.QSplitter()
        self.horizontal_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.horizontal_splitter.addWidget(self.vertical_splitter_left)
        self.horizontal_splitter.addWidget(self.vertical_splitter)
        # self.horizontal_splitter.addWidget(self.vertical_splitter)

        self._layout.addWidget(self.horizontal_splitter, 10)
        self._layout.addWidget(self.integration_status_widget, 0)
        self.setLayout(self._layout)

        self.create_shortcuts()

        self.bkg_image_scale_sb.setKeyboardTracking(False)
        self.bkg_image_offset_sb.setKeyboardTracking(False)

        self.qa_bkg_pattern_inspect_btn.setVisible(False)

        self.mask_transparent_cb.setVisible(False)

        self.file_info_widget = FileInfoWidget(self)
        self.move_widget = MoveStageWidget(self)
        # self.map_2D_widget = Map2DWidget(self)
        self.batch_widget = BatchWidget(self)

        self.img_frame_size = QtCore.QSize(400, 500)
        self.img_frame_position = QtCore.QPoint(0, 0)

        self.img_mode = 'Image'

    def create_shortcuts(self):
        img_control_widget = self.integration_control_widget.img_control_widget.file_widget
        self.image_control_widget = img_control_widget
        self.load_img_btn = img_control_widget.load_btn
        self.autoprocess_cb = img_control_widget.file_cb
        self.img_step_file_widget = img_control_widget.step_file_widget
        self.img_step_series_widget = img_control_widget.step_series_widget
        self.img_filename_txt = img_control_widget.file_txt
        self.img_directory_txt = img_control_widget.directory_txt
        self.img_directory_btn = img_control_widget.directory_btn
        self.file_info_btn = self.integration_control_widget.img_control_widget.file_info_btn
        self.move_btn = self.integration_control_widget.img_control_widget.move_btn
        self.move_widget_btn = self.integration_control_widget.img_control_widget.move_btn
        self.img_batch_mode_integrate_rb = self.integration_control_widget.img_control_widget.batch_mode_integrate_rb
        self.img_batch_mode_add_rb = self.integration_control_widget.img_control_widget.batch_mode_add_rb
        self.img_batch_mode_average_rb = self.integration_control_widget.img_control_widget.batch_mode_average_rb
        self.img_batch_mode_image_save_rb = self.integration_control_widget.img_control_widget.batch_mode_image_save_rb

        pattern_file_widget = self.integration_control_widget.pattern_control_widget.file_widget
        self.pattern_load_btn = pattern_file_widget.load_btn
        self.pattern_autocreate_cb = pattern_file_widget.file_cb
        self.pattern_previous_btn = pattern_file_widget.step_file_widget.previous_btn
        self.pattern_next_btn = pattern_file_widget.step_file_widget.next_btn
        self.pattern_browse_step_txt = pattern_file_widget.step_file_widget.step_txt
        self.pattern_browse_by_name_rb = pattern_file_widget.step_file_widget.browse_by_name_rb
        self.pattern_browse_by_time_rb = pattern_file_widget.step_file_widget.browse_by_time_rb
        self.pattern_filename_txt = pattern_file_widget.file_txt
        self.pattern_directory_txt = pattern_file_widget.directory_txt
        self.pattern_directory_btn = pattern_file_widget.directory_btn
        self.pattern_header_xy_cb = self.integration_control_widget.pattern_control_widget.xy_cb
        self.pattern_header_chi_cb = self.integration_control_widget.pattern_control_widget.chi_cb
        self.pattern_header_dat_cb = self.integration_control_widget.pattern_control_widget.dat_cb
        self.pattern_header_fxye_cb = self.integration_control_widget.pattern_control_widget.fxye_cb
        self.pattern_headers = []
        self.pattern_headers.append(self.pattern_header_xy_cb)
        self.pattern_headers.append(self.pattern_header_chi_cb)
        self.pattern_headers.append(self.pattern_header_dat_cb)
        self.pattern_headers.append(self.pattern_header_fxye_cb)

        phase_control_widget = self.integration_control_widget.phase_control_widget
        self.phase_widget = phase_control_widget
        self.phase_add_btn = phase_control_widget.add_btn
        self.phase_edit_btn = phase_control_widget.edit_btn
        self.phase_del_btn = phase_control_widget.delete_btn
        self.phase_clear_btn = phase_control_widget.clear_btn
        self.phase_save_list_btn = phase_control_widget.save_list_btn
        self.phase_load_list_btn = phase_control_widget.load_list_btn
        self.phase_tw = phase_control_widget.phase_tw
        self.phase_pressure_step_msb = phase_control_widget.pressure_step_msb
        self.phase_temperature_step_msb = phase_control_widget.temperature_step_msb
        self.phase_apply_to_all_cb = phase_control_widget.apply_to_all_cb

        overlay_control_widget = self.integration_control_widget.overlay_control_widget
        self.overlay_widget = overlay_control_widget
        self.overlay_add_btn = overlay_control_widget.add_btn
        self.overlay_del_btn = overlay_control_widget.delete_btn
        self.overlay_clear_btn = overlay_control_widget.clear_btn
        self.overlay_move_up_btn = overlay_control_widget.move_up_btn
        self.overlay_move_down_btn = overlay_control_widget.move_down_btn
        self.overlay_tw = overlay_control_widget.overlay_tw
        self.overlay_scale_step_msb = overlay_control_widget.scale_step_msb
        self.overlay_offset_step_msb = overlay_control_widget.offset_step_msb
        self.waterfall_separation_msb = overlay_control_widget.waterfall_separation_msb
        self.waterfall_btn = overlay_control_widget.waterfall_btn
        self.reset_waterfall_btn = overlay_control_widget.waterfall_reset_btn
        self.overlay_set_as_bkg_btn = overlay_control_widget.set_as_bkg_btn

        corrections_control_widget = self.integration_control_widget.corrections_control_widget
        self.cbn_groupbox = corrections_control_widget.cbn_seat_gb
        self.cbn_param_tw = corrections_control_widget.cbn_param_tw
        self.cbn_plot_btn = corrections_control_widget.cbn_seat_plot_btn
        self.oiadac_groupbox = corrections_control_widget.oiadac_gb
        self.oiadac_param_tw = corrections_control_widget.oiadac_param_tw
        self.oiadac_plot_btn = corrections_control_widget.oiadac_plot_btn
        self.transfer_gb = corrections_control_widget.transfer_gb
        self.transfer_load_original_btn = corrections_control_widget.transfer_load_original_btn
        self.transfer_load_response_btn = corrections_control_widget.transfer_load_response_btn
        self.transfer_plot_btn = corrections_control_widget.transfer_plot_btn
        self.transfer_original_filename_lbl = corrections_control_widget.transfer_original_filename_lbl
        self.transfer_response_filename_lbl = corrections_control_widget.transfer_response_filename_lbl

        background_control_widget = self.integration_control_widget.background_control_widget
        self.bkg_image_load_btn = background_control_widget.load_image_btn
        self.bkg_image_filename_lbl = background_control_widget.filename_lbl
        self.bkg_image_delete_btn = background_control_widget.remove_image_btn
        self.bkg_image_scale_sb = background_control_widget.scale_sb
        self.bkg_image_scale_step_msb = background_control_widget.scale_step_msb
        self.bkg_image_offset_sb = background_control_widget.offset_sb
        self.bkg_image_offset_step_msb = background_control_widget.offset_step_msb
        self.bkg_pattern_gb = background_control_widget.pattern_background_gb
        self.bkg_pattern_smooth_width_sb = background_control_widget.smooth_with_sb
        self.bkg_pattern_iterations_sb = background_control_widget.iterations_sb
        self.bkg_pattern_poly_order_sb = background_control_widget.poly_order_sb
        self.bkg_pattern_x_min_txt = background_control_widget.x_range_min_txt
        self.bkg_pattern_x_max_txt = background_control_widget.x_range_max_txt
        self.bkg_pattern_inspect_btn = background_control_widget.inspect_btn
        self.bkg_pattern_save_btn = background_control_widget.save_btn
        self.bkg_pattern_as_overlay_btn = background_control_widget.as_overlay

        options_control_widget = self.integration_control_widget.integration_options_widget
        self.bin_count_txt = options_control_widget.bin_count_txt
        self.automatic_binning_cb = options_control_widget.bin_count_cb
        self.correct_solid_angle_cb = options_control_widget.correct_solid_angle_cb
        self.supersampling_sb = options_control_widget.supersampling_sb
        self.oned_full_range_btn = options_control_widget.oned_full_toggle_btn
        self.oned_azimuth_min_txt = options_control_widget.oned_azimuth_min_txt
        self.oned_azimuth_max_txt = options_control_widget.oned_azimuth_max_txt

        self.mouse_x_lbl = self.integration_status_widget.mouse_pos_widget.cur_pos_widget.x_pos_lbl
        self.mouse_y_lbl = self.integration_status_widget.mouse_pos_widget.cur_pos_widget.y_pos_lbl
        self.mouse_int_lbl = self.integration_status_widget.mouse_pos_widget.cur_pos_widget.int_lbl
        self.click_x_lbl = self.integration_status_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl
        self.click_y_lbl = self.integration_status_widget.mouse_pos_widget.clicked_pos_widget.y_pos_lbl
        self.click_int_lbl = self.integration_status_widget.mouse_pos_widget.clicked_pos_widget.int_lbl
        self.mouse_tth_lbl = self.integration_status_widget.mouse_unit_widget.cur_unit_widget.tth_lbl
        self.mouse_q_lbl = self.integration_status_widget.mouse_unit_widget.cur_unit_widget.q_lbl
        self.mouse_d_lbl = self.integration_status_widget.mouse_unit_widget.cur_unit_widget.d_lbl
        self.mouse_azi_lbl = self.integration_status_widget.mouse_unit_widget.cur_unit_widget.azi_lbl
        self.click_tth_lbl = self.integration_status_widget.mouse_unit_widget.clicked_unit_widget.tth_lbl
        self.click_q_lbl = self.integration_status_widget.mouse_unit_widget.clicked_unit_widget.q_lbl
        self.click_d_lbl = self.integration_status_widget.mouse_unit_widget.clicked_unit_widget.d_lbl
        self.click_azi_lbl = self.integration_status_widget.mouse_unit_widget.clicked_unit_widget.azi_lbl
        self.bkg_name_lbl = self.integration_status_widget.bkg_name_lbl

        pattern_widget = self.integration_pattern_widget
        self.qa_save_pattern_btn = pattern_widget.save_pattern_btn
        self.qa_set_as_overlay_btn = pattern_widget.as_overlay_btn
        self.qa_set_as_background_btn = pattern_widget.as_bkg_btn
        self.load_calibration_btn = pattern_widget.load_calibration_btn
        self.calibration_lbl = pattern_widget.calibration_lbl
        self.pattern_tth_btn = pattern_widget.tth_btn
        self.pattern_q_btn = pattern_widget.q_btn
        self.pattern_d_btn = pattern_widget.d_btn
        self.qa_bkg_pattern_btn = pattern_widget.background_btn
        self.qa_bkg_pattern_inspect_btn = pattern_widget.background_inspect_btn
        self.set_wavelnegth_btn = pattern_widget.set_wavelength_btn
        self.wavelength_lbl = pattern_widget.wavelength_lbl
        self.qa_bkg_pattern_btn = pattern_widget.background_btn
        self.qa_bkg_pattern_inspect_btn = pattern_widget.background_inspect_btn
        self.antialias_btn = pattern_widget.antialias_btn
        self.pattern_auto_range_btn = pattern_widget.auto_range_btn
        self.pattern_widget = pattern_widget.pattern_view

        image_widget = self.integration_image_widget
        self.qa_save_img_btn = image_widget.save_image_btn
        self.img_frame = image_widget
        self.img_roi_btn = image_widget.roi_btn
        self.img_mode_btn = image_widget.mode_btn
        self.img_mask_btn = image_widget.mask_btn
        self.img_phases_btn = image_widget.phases_btn
        self.cake_shift_azimuth_sl = image_widget.cake_shift_azimuth_sl
        self.mask_transparent_cb = image_widget.transparent_cb
        self.img_autoscale_btn = image_widget.autoscale_btn
        self.img_dock_btn = image_widget.undock_btn
        self.img_widget = image_widget.img_view
        self.cake_widget = image_widget.cake_view
        self.img_show_background_subtracted_btn = image_widget.show_background_subtracted_img_btn

        self.frame_img_positions_widget = self.integration_image_widget.position_and_unit_widget
        self.tabWidget = self.integration_control_widget
        self.img_widget_mouse_x_lbl = self.integration_image_widget.mouse_pos_widget.cur_pos_widget.x_pos_lbl
        self.img_widget_mouse_y_lbl = self.integration_image_widget.mouse_pos_widget.cur_pos_widget.y_pos_lbl
        self.img_widget_mouse_int_lbl = self.integration_image_widget.mouse_pos_widget.cur_pos_widget.int_lbl
        self.img_widget_click_x_lbl = self.integration_image_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl
        self.img_widget_click_y_lbl = self.integration_image_widget.mouse_pos_widget.clicked_pos_widget.y_pos_lbl
        self.img_widget_click_int_lbl = self.integration_image_widget.mouse_pos_widget.clicked_pos_widget.int_lbl
        self.img_widget_mouse_tth_lbl = self.integration_image_widget.mouse_unit_widget.cur_unit_widget.tth_lbl
        self.img_widget_mouse_q_lbl = self.integration_image_widget.mouse_unit_widget.cur_unit_widget.q_lbl
        self.img_widget_mouse_d_lbl = self.integration_image_widget.mouse_unit_widget.cur_unit_widget.d_lbl
        self.img_widget_mouse_azi_lbl = self.integration_image_widget.mouse_unit_widget.cur_unit_widget.azi_lbl
        self.img_widget_click_tth_lbl = self.integration_image_widget.mouse_unit_widget.clicked_unit_widget.tth_lbl
        self.img_widget_click_q_lbl = self.integration_image_widget.mouse_unit_widget.clicked_unit_widget.q_lbl
        self.img_widget_click_d_lbl = self.integration_image_widget.mouse_unit_widget.clicked_unit_widget.d_lbl
        self.img_widget_click_azi_lbl = self.integration_image_widget.mouse_unit_widget.clicked_unit_widget.azi_lbl

        self.footer_img_mouse_position_widget = self.integration_status_widget.mouse_pos_widget
        self.change_view_btn = self.integration_status_widget.change_view_btn

    def switch_to_cake(self):
        self.img_widget.img_view_box.setAspectLocked(False)
        self.img_widget.activate_vertical_line()

    def switch_to_img(self):
        self.img_widget.img_view_box.setAspectLocked(True)
        self.img_widget.deactivate_vertical_line()

    def dock_img(self, bool_value):
        if not bool_value:
            self.img_dock_btn.setText('Dock')

            # save current splitter state
            self.horizontal_splitter_state = self.horizontal_splitter.saveState()
            self.vertical_splitter_left_state = self.vertical_splitter_left.saveState()

            # splitter_handle = self.horizontal_splitter.handle(1)
            # splitter_handle.setEnabled(False)

            self.img_frame.setParent(self)
            self.img_frame.setWindowFlags(QtCore.Qt.Window | QtCore.Qt.CustomizeWindowHint | \
                                          QtCore.Qt.WindowMinimizeButtonHint | QtCore.Qt.WindowMaximizeButtonHint)
            self.frame_img_positions_widget.show()
            self.img_frame.resize(self.img_frame_size)
            self.img_frame.move(self.img_frame_position)
            self.footer_img_mouse_position_widget.hide()
            self.img_frame.show()
        elif bool_value:
            self.img_dock_btn.setText('Undock')

            # save the current position and size of the img_frame to be able to restore it later
            self.img_frame_size = self.img_frame.size()
            self.img_frame_position = self.img_frame.pos()

            # reassign visibilities of mouse position and click labels
            self.footer_img_mouse_position_widget.show()
            self.frame_img_positions_widget.hide()

            # remove all widgets/frames from horizontal splitter to be able to arrange them in the correct order
            self.img_frame.setParent(self.vertical_splitter_left)

            self.vertical_splitter_left.insertWidget(1, self.img_frame)
            # self.horizontal_splitter.addWidget(self.vertical_splitter)

            # restore the previously used size when image was undocked
            self.horizontal_splitter.restoreState(self.horizontal_splitter_state)
            self.vertical_splitter_left.restoreState(self.vertical_splitter_left_state)

    def get_progress_dialog(self, message, abort_text, num_points):
        progress_dialog = QtWidgets.QProgressDialog(message, abort_text, 0,
                                                    num_points, self)
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
        progress_dialog.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        progress_dialog.move(
            int(self.pattern_widget.pg_layout.x() + self.pattern_widget.pg_layout.size().width() / 2.0 -
                progress_dialog.size().width() / 2.0),
            int(self.pattern_widget.pg_layout.y() + self.pattern_widget.pg_layout.size().height() / 2.0 -
                progress_dialog.size().height() / 2.0))
        progress_dialog.show()
        return progress_dialog

    def show_error_msg(self, msg):
        msg_box = QtWidgets.QMessageBox(self)
        msg_box.setWindowFlags(QtCore.Qt.Tool)
        msg_box.setText(msg)
        msg_box.setIcon(QtWidgets.QMessageBox.Critical)
        msg_box.setWindowTitle('Error')
        msg_box.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg_box.setDefaultButton(QtWidgets.QMessageBox.Ok)
        msg_box.exec_()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets

from ...CustomWidgets import LabelAlignRight
from ..CustomWidgets import BrowseFileWidget


class PatternWidget(QtWidgets.QWidget):
    def __init__(self):
        super(PatternWidget, self).__init__()

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(5)

        self.file_widget = BrowseFileWidget(files='Pattern', checkbox_text='autocreate')

        self._layout.addWidget(self.file_widget)

        self.xy_cb = QtWidgets.QCheckBox('.xy')
        self.xy_cb.setChecked(True)
        self.chi_cb = QtWidgets.QCheckBox('.chi')
        self.dat_cb = QtWidgets.QCheckBox('.dat')
        self.fxye_cb = QtWidgets.QCheckBox('.fxye')
        self._pattern_types_layout = QtWidgets.QHBoxLayout()
        self._pattern_types_layout.addWidget(LabelAlignRight('Pattern types:'))
        self._pattern_types_layout.addWidget(self.xy_cb)
        self._pattern_types_layout.addWidget(self.chi_cb)
        self._pattern_types_layout.addWidget(self.dat_cb)
        self._pattern_types_layout.addWidget(self.fxye_cb)
        self._pattern_types_layout.addStretch()

        self._layout.addLayout(self._pattern_types_layout)
        self._layout.addStretch()
        self.setLayout(self._layout)

        self.setToolTips()

    def setToolTips(self):
        self.xy_cb.setToolTip('Create .xy files')
        self.chi_cb.setToolTip('Create .chi files')
        self.dat_cb.setToolTip('Create .dat files')
        self.fxye_cb.setToolTip('Create .fxye files')
        self.file_widget.file_cb.setToolTip('Autocreate patterns for each loaded image')


# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from ..widgets.UtilityWidgets import save_file_dialog

# imports for type hinting in PyCharm -- DO NOT DELETE
from ..widgets.ConfigurationWidget import ConfigurationWidget
from ..model.DioptasModel import DioptasModel


class ConfigurationController(object):
    """
    Deals with all the signal handling and model upgrades related to be using multiple configurations.
    """

    def __init__(self, configuration_widget, dioptas_model, controllers=()):
        """
        :type configuration_widget: ConfigurationWidget
        :type dioptas_model: DioptasModel
        """
        self.widget = configuration_widget
        self.model = dioptas_model
        self.controllers = controllers

        self.update_configuration_widget()

        self.create_signals()

    def create_signals(self):
        self.widget.add_configuration_btn.clicked.connect(self.model.add_configuration)
        self.widget.remove_configuration_btn.clicked.connect(self.model.remove_configuration)

        self.widget.configuration_selected.connect(self.model.select_configuration)

        self.model.configuration_added.connect(self.update_configuration_widget)
        self.model.configuration_removed.connect(self.update_configuration_widget)
        self.model.configuration_selected.connect(self.configuration_selected)

        self.widget.next_file_btn.clicked.connect(self.load_next_file)
        self.widget.previous_file_btn.clicked.connect(self.load_previous_file)

        self.widget.next_folder_btn.clicked.connect(self.load_next_folder)
        self.widget.previous_folder_btn.clicked.connect(self.load_previous_folder)

        self.widget.factor_txt.editingFinished.connect(self.factor_txt_changed)

        self.widget.combine_patterns_btn.clicked.connect(self.combine_patterns_btn_clicked)
        self.widget.saved_combined_patterns_btn.clicked.connect(self.save_combined_patterns_btn_clicked)
        self.widget.combine_cakes_btn.clicked.connect(self.combine_cakes_btn_clicked)

    def update_configuration_widget(self):
        self.widget.update_configuration_btns(
            configurations=self.model.configurations,
            cur_ind=self.model.configuration_ind
        )

    def configuration_selected(self):
        self.widget.factor_txt.setText(str(self.model.img_model.factor))

    def combine_patterns_btn_clicked(self):
        self.model.combine_patterns = self.widget.combine_patterns_btn.isChecked()

    def save_combined_patterns_btn_clicked(self):
        img_filename, _ = os.path.splitext(os.path.basename(self.model.img_model.filename))
        filename = save_file_dialog(
            self.widget, "Save Combined Pattern Data.",
            os.path.join(self.model.working_directories['pattern'],
                         img_filename + '_combined.dat'), ('.dat'))

        if filename != '':
            self.model.save_combined_pattern(filename)

    def combine_cakes_btn_clicked(self):
        self.model.combine_cakes = self.widget.combine_cakes_btn.isChecked()

    def factor_txt_changed(self):
        self.model.img_model.factor = float(str(self.widget.factor_txt.text()))

    def load_next_file(self):
        pos = int(str(self.widget.file_iterator_pos_txt.text()))
        self.model.next_image(pos)

    def load_previous_file(self):
        pos = int(str(self.widget.file_iterator_pos_txt.text()))
        self.model.previous_image(pos)

    def load_next_folder(self):
        self.model.next_folder(mec_mode=bool(self.widget.mec_cb.isChecked()))

    def load_previous_folder(self):
        self.model.previous_folder(mec_mode=bool(self.widget.mec_cb.isChecked()))



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import itertools
from math import degrees
from urllib.request import pathname2url

from CifFile import ReadCif
from .jcpds import jcpds
from ... import data_path

import numpy as np
import json

with open(os.path.join(data_path, "atomic_scattering_params.json")) as f:
    ATOMIC_SCATTERING_PARAMS = json.load(f)

with open(os.path.join(data_path, "periodic_table.json")) as f:
    PERIODIC_TABLE = json.load(f)


class CifConverter(object):
    # Tolerance in which to treat two peaks as having the same two theta.
    TWO_THETA_TOL = 1e-5

    def __init__(self, wavelength, min_d_spacing=0.5, min_intensity=0.5):
        """
        Calculates the x-ray diffraction intensities for a specific cif phase.
        :param wavelength: Wavelength for the two theta calculation
        :param min_d_spacing: all calculated reflections will have a d-spacing above this value
        :param min_intensity: all calculated reflections will have an intensity greater than this value
        :return:
        """
        self.wavelength = wavelength
        self.min_d_spacing = min_d_spacing
        self.min_intensity = min_intensity

    def convert_cif_to_jcpds(self, filename):
        """
        Reads a cif file and returns a jcpds with correct reflection and intensities
        :param filename:  cif filename
        :return: converted jcpds object
        :rtype: jcpds
        """
        file_url = 'file:' + pathname2url(filename)
        cif_file = ReadCif(file_url)
        cif_phase = CifPhase(cif_file[cif_file.keys()[0]])
        jcpds_phase = self.convert_cif_phase_to_jcpds(cif_phase)
        jcpds_phase.filename = filename
        jcpds_phase.name = os.path.splitext(os.path.basename(filename))[0]
        jcpds_phase.params['modified'] = False

        return jcpds_phase

    def convert_cif_phase_to_jcpds(self, cif_phase):
        """
        Converts a CifPhase into a jcpds object by calculating the intensities and multiplicities for all reflections
        :param cif_phase: input CifPhase
        :type cif_phase: CifPhase
        :return: converted jcpds object
        :rtype: jcpds
        """
        reflections = self._calculate_hkl_within_sphere_and_min_d_spacing(cif_phase)
        xrd_reflections = self._calculate_reflection_intensities(cif_phase, reflections)
        jcpds_phase = self._create_jcpds_from_cif_parameters(cif_phase)

        for reflection in xrd_reflections:
            jcpds_phase.add_reflection(reflection.h,
                                       reflection.k,
                                       reflection.l,
                                       reflection.intensity,
                                       reflection.d_spacing)
        return jcpds_phase

    def _create_jcpds_from_cif_parameters(self, cif_phase):
        """
        Creates a jcpds object from a cif_phase using cell parameters, symmetries, and file information. Does not
        calculate intensities for any reflection e.g. the jcpds will have 0 reflections
        :param cif_phase:
        :type cif_phase: CifPhase
        :return:
        """
        jcpds_phase = jcpds()

        jcpds_phase.params['a0'] = cif_phase.a
        jcpds_phase.params['b0'] = cif_phase.b
        jcpds_phase.params['c0'] = cif_phase.c
        jcpds_phase.params['alpha0'] = cif_phase.alpha
        jcpds_phase.params['beta0'] = cif_phase.beta
        jcpds_phase.params['gamma0'] = cif_phase.gamma
        jcpds_phase.params['v0'] = cif_phase.volume
        jcpds_phase.params['symmetry'] = cif_phase.symmetry
        jcpds_phase.params['comments'] = [cif_phase.comments]

        return jcpds_phase

    def _calculate_hkl_within_sphere_and_min_d_spacing(self, cif_phase):
        """
        Generates a list of hkl reflections which can satisfy the diffraction condition using the given wavelength and
        also the minimum d spacing
        :return: list of reflections
        :rtype: list[Reflection]
        """
        max_h = np.floor(2 * cif_phase.a / self.wavelength)
        max_k = np.floor(2 * cif_phase.b / self.wavelength)
        max_l = np.floor(2 * cif_phase.c / self.wavelength)

        h = reversed(np.arange(-max_h + 1, max_h, 1))
        k = reversed(np.arange(-max_k + 1, max_k, 1))
        l = reversed(np.arange(-max_l + 1, max_l, 1))
        s = [h, k, l]

        reciprocal_hkl = np.array(list(itertools.product(*s)))
        h = reciprocal_hkl[:, 0]
        k = reciprocal_hkl[:, 1]
        l = reciprocal_hkl[:, 2]

        d_hkl = compute_d_hkl(h, k, l, cif_phase)
        good_indices = np.argwhere(d_hkl > self.min_d_spacing)

        reflections = []
        for i in good_indices:
            reflections.append(Reflection(int(h[i][0]), int(k[i][0]), int(l[i][0]), d_hkl[i][0]))

        return reflections

    def _calculate_reflection_intensities(self, cif_phase, base_reflections):
        """
        :param cif_phase:
        :param base_reflections:
        :return:
        """
        # provide atom parameters as lists
        atom_numbers = []
        form_coefficients = []
        fractional_coordinates = []
        occupancies = []
        for atom in cif_phase.atoms:
            atom_numbers.append(PERIODIC_TABLE[atom[0]]['Atomic no'])
            try:
                c = ATOMIC_SCATTERING_PARAMS[atom[0]]
            except KeyError:
                raise ValueError("Unable to calculate XRD pattern as "
                                 "there is no scattering coefficients for"
                                 " %s." % atom[0])
            form_coefficients.append(c)
            fractional_coordinates.append(atom[1:4])
            occupancies.append(atom[4])

        atom_numbers = np.array(atom_numbers)
        form_coefficients = np.array(form_coefficients)
        fractional_coordinates = np.array(fractional_coordinates)
        occupancies = np.array(occupancies)

        two_thetas = []
        peaks = {}

        for reflection in base_reflections:
            d_hkl = reflection.d_spacing
            g_hkl = 1. / d_hkl
            if g_hkl != 0:
                hkl = np.array([reflection.h, reflection.k, reflection.l])

                s = g_hkl * 0.5
                s2 = s ** 2

                theta = np.arcsin(self.wavelength * g_hkl * 0.5)
                two_theta = degrees(2 * theta)

                g_dot_r = np.dot(fractional_coordinates, np.transpose([hkl])).T[0]

                fs = atom_numbers - 41.78214 * s2 * np.sum(
                    form_coefficients[:, :, 0] * np.exp(-form_coefficients[:, :, 1] * s2), axis=1)
                f_hkl = np.sum(fs * occupancies * np.exp(2j * np.pi * g_dot_r))
                i_hkl = (f_hkl * f_hkl.conjugate()).real

                ind = np.where(np.abs(np.subtract(two_thetas, two_theta)) < CifConverter.TWO_THETA_TOL)

                lorentz_factor = (1 + np.cos(2 * theta) ** 2) / (np.sin(theta) ** 2 * np.cos(theta))

                if len(ind[0]) > 0:
                    peak_ind = two_thetas[ind[0][0]]
                    peaks[peak_ind][0] += i_hkl * lorentz_factor
                    peaks[peak_ind][1].append(tuple(hkl))
                else:
                    peaks[two_theta] = [i_hkl * lorentz_factor, [tuple(hkl)], d_hkl]
                    two_thetas.append(two_theta)

        max_intensity = max([v[0] for v in peaks.values()])
        calculated_reflections = []
        for k in sorted(peaks.keys()):
            v = peaks[k]
            scaled_intensity = v[0] / max_intensity * 100
            fam = get_unique_families(v[1])
            if scaled_intensity > self.min_intensity:
                calculated_reflections.append(
                    Reflection(
                        *list(fam.keys())[0],
                        d_spacing=v[2],
                        intensity=scaled_intensity,
                        multiplicity=fam[list(fam.keys())[0]]
                    )
                )

        return calculated_reflections


class Reflection:
    def __init__(self, h, k, l, d_spacing, intensity=None, multiplicity=1):
        self.h = h
        self.k = k
        self.l = l
        self.multiplicity = multiplicity
        self.d_spacing = d_spacing
        self.intensity = intensity

    def __repr__(self):
        return "({},{},{}) x {} {:.5f} - {}".format(self.h, self.k, self.l, self.multiplicity,
                                                    self.d_spacing, self.intensity)


def compute_d_hkl(h, k, l, cif_phase):
    a = cif_phase.a
    b = cif_phase.b
    c = cif_phase.c
    alpha = np.deg2rad(cif_phase.alpha)
    beta = np.deg2rad(cif_phase.beta)
    gamma = np.deg2rad(cif_phase.gamma)

    symmetry = cif_phase.symmetry

    if symmetry == 'CUBIC':
        d2inv = (h ** 2 + k ** 2 + l ** 2) / a ** 2
    elif symmetry == 'TETRAGONAL':
        d2inv = (h ** 2 + k ** 2) / a ** 2 + l ** 2 / c ** 2
    elif symmetry == 'ORTHORHOMBIC':
        d2inv = h ** 2 / a ** 2 + k ** 2 / b ** 2 + l ** 2 / c ** 2
    elif symmetry == 'HEXAGONAL' or symmetry == 'TRIGONAL':
        d2inv = (h ** 2 + h * k + k ** 2) * 4. / 3. / a ** 2 + l ** 2 / c ** 2
    elif symmetry == 'RHOMBOHEDRAL':
        d2inv = (((1. + np.cos(alpha)) * ((h ** 2 + k ** 2 + l ** 2) -
                                          (1 - np.tan(0.5 * alpha) ** 2) * (h * k + k * l + l * h))) /
                 (a ** 2 * (1 + np.cos(alpha) - 2 * np.cos(alpha) ** 2)))
    elif symmetry == 'MONOCLINIC':
        d2inv = (h ** 2 / np.sin(beta) ** 2 / a ** 2 +
                 k ** 2 / b ** 2 +
                 l ** 2 / np.sin(beta) ** 2 / c ** 2 +
                 2 * h * l * np.cos(beta) / (a * c * np.sin(beta) ** 2))
    elif symmetry == 'TRICLINIC':
        V = (a * b * c *
             np.sqrt(1. - np.cos(alpha) ** 2 - np.cos(beta) ** 2 -
                     np.cos(gamma) ** 2 +
                     2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma)))
        s11 = b ** 2 * c ** 2 * np.sin(alpha) ** 2
        s22 = a ** 2 * c ** 2 * np.sin(beta) ** 2
        s33 = a ** 2 * b ** 2 * np.sin(gamma) ** 2
        s12 = a * b * c ** 2 * (np.cos(alpha) * np.cos(beta) -
                                np.cos(gamma))
        s23 = a ** 2 * b * c * (np.cos(beta) * np.cos(gamma) -
                                np.cos(alpha))
        s31 = a * b ** 2 * c * (np.cos(gamma) * np.cos(alpha) -
                                np.cos(beta))
        d2inv = (s11 * h ** 2 + s22 * k ** 2 + s33 * l ** 2 +
                 2. * s12 * h * k + 2. * s23 * k * l + 2. * s31 * l * h) / V ** 2
    d_spacings = np.sqrt(1. / d2inv)

    return d_spacings


def get_unique_families(hkls):
    """
    Returns unique families of Miller indices. Families must be permutations
    of each other.

    Args:
        hkls ([h, k, l]): List of Miller indices.

    Returns:
        {hkl: multiplicity}: A dict with unique hkl and multiplicity.
    """

    # TODO: Definitely can be sped up.
    def is_perm(hkl1, hkl2):
        h1 = map(abs, hkl1)
        h2 = map(abs, hkl2)
        return all([i == j for i, j in zip(sorted(h1), sorted(h2))])

    unique = {}
    for hkl1 in hkls:
        found = False
        for hkl2 in unique.keys():
            if is_perm(hkl1, hkl2):
                found = True
                unique[hkl2] += 1
                break
        if not found:
            unique[hkl1] = 1

    return unique


class CifPhase(object):
    """
    Phase created from a cif dictionary (a cif file can have several phases which can be read as a list by PyCifRw)

    This automatically creates symmetry equivalent position of atoms for further processing of the file.
    """

    def __init__(self, cif_dictionary):
        """

        :param cif_dictionary:
        :return:
        """
        self.cif_dictionary = cif_dictionary

        self.a = convert_cif_number_to_float(cif_dictionary['_cell_length_a'])
        self.b = convert_cif_number_to_float(cif_dictionary['_cell_length_b'])
        self.c = convert_cif_number_to_float(cif_dictionary['_cell_length_c'])

        self.alpha = convert_cif_number_to_float(cif_dictionary['_cell_angle_alpha'])
        self.beta = convert_cif_number_to_float(cif_dictionary['_cell_angle_beta'])
        self.gamma = convert_cif_number_to_float(cif_dictionary['_cell_angle_gamma'])

        if '_cell_volume' in cif_dictionary.keys():
            self.volume = convert_cif_number_to_float(cif_dictionary['_cell_volume'])
        else:
            self.volume = calculate_cell_volume(self.a, self.b, self.c,
                                                np.deg2rad(self.alpha), np.deg2rad(self.beta), np.deg2rad(self.gamma))

        if '_symmetry_space_group_name_h-m' in cif_dictionary.keys():
            self.space_group = cif_dictionary['_symmetry_space_group_name_h-m']
        elif '_symmetry_space_group_name_h-m_alt' in cif_dictionary.keys():
            self.space_group = cif_dictionary['_symmetry_space_group_name_h-m_alt']
        else:
            self.space_group = None

        if '_symmetry_Int_Tables_number'.lower() in cif_dictionary.keys():
            self.space_group_number = cif_dictionary.get('_symmetry_Int_Tables_number'.lower())
        elif '_space_group_IT_number'.lower() in cif_dictionary.keys():
            self.space_group_number = cif_dictionary.get('_space_group_IT_number'.lower())
        else:
            self.space_group_number = None

        if self.space_group_number is not None:
            self.space_group_number = int(self.space_group_number)

        self.symmetry = self.get_symmetry_from_space_group_number(self.space_group_number)

        self.comments = ''
        self.read_file_information()

        if '_symmetry_equiv_pos_as_xyz' in cif_dictionary.keys():
            self.symmetry_operations = cif_dictionary['_symmetry_equiv_pos_as_xyz']
        elif '_space_group_symop_operation_xyz' in cif_dictionary.keys():
            self.symmetry_operations = cif_dictionary['_space_group_symop_operation_xyz']

        for i in range(len(self.symmetry_operations)):
            self.symmetry_operations[i] = self.symmetry_operations[i].replace("/", "./")

        self._atom_labels = cif_dictionary['_atom_site_label']
        self._atom_x = [float(convert_cif_number_to_float(s)) for s in cif_dictionary['_atom_site_fract_x']]
        self._atom_y = [float(convert_cif_number_to_float(s)) for s in cif_dictionary['_atom_site_fract_y']]
        self._atom_z = [float(convert_cif_number_to_float(s)) for s in cif_dictionary['_atom_site_fract_z']]
        if '_atom_site_occupancy' in cif_dictionary.keys():
            self._atom_occupancy = [float(convert_cif_number_to_float(s)) for s in
                                    cif_dictionary['_atom_site_occupancy']]
        else:
            self._atom_occupancy = [1] * len(self._atom_labels)

        # Create a list of 4-tuples, where each tuple is an atom:
        # [ ('Si', 0.4697, 0.0, 0.0),  ('O', 0.4135, 0.2669, 0.1191),  ... ]
        self.atoms = [(self._atom_labels[i], self._atom_x[i], self._atom_y[i], self._atom_z[i],
                       self._atom_occupancy[i]) for i in range(len(self._atom_labels))]
        self.clean_atoms()
        self.generate_symmetry_equivalents()

    def clean_atoms(self):
        """
        Cleaning all atom labels and check atom positions
        """
        for i in range(len(self.atoms)):
            (name, xn, yn, zn, occu) = self.atoms[i]
            xn = (xn + 10.0) % 1.0
            yn = (yn + 10.0) % 1.0
            zn = (zn + 10.0) % 1.0
            name = self.convert_element(name)
            self.atoms[i] = (name, xn, yn, zn, occu)

    def generate_symmetry_equivalents(self):
        # The CIF file consists of a few atom positions plus several "symmetry
        # operations" that indicate the other atom positions within the unit cell.  So
        # using these operations, create copies of the atoms until no new copies can be
        # made.

        # Two atoms are on top of each other if they are less than "eps" away.
        eps = 0.0001  # in Angstrom

        # For each atom, apply each symmetry operation to create a new atom.
        i = 0
        while (i < len(self.atoms)):

            label, x, y, z, occu = self.atoms[i]

            for op in self.symmetry_operations:

                # Python is awesome: calling e.g. eval('x,y,1./2+z') will convert the
                # string into a 3-tuple using the current values for x,y,z!
                xn, yn, zn = eval(op)

                # Make sure that the new atom lies within the unit cell.
                xn = (xn + 10.0) % 1.0
                yn = (yn + 10.0) % 1.0
                zn = (zn + 10.0) % 1.0

                # Check if the new position is actually new, or the same as a previous
                # atom.
                new_atom = True
                for at in self.atoms:
                    if (abs(at[1] - xn) < eps and abs(at[2] - yn) < eps and abs(at[3] - zn) < eps) \
                            and at[0] == label:
                        new_atom = False

                # If the atom is new, add it to the list!
                if (new_atom):
                    self.atoms.append((label, xn, yn, zn, occu))  # add a 4-tuple

            # Update the loop iterator.
            i += 1

        # Sort the atoms according to type alphabetically.
        self.atoms = sorted(self.atoms, key=lambda at: at[0])
        self.atoms.reverse()

    def convert_element(self, label):
        """
        Convert atom labels such as 'Oa1' into 'O'.
        """

        elem2 = ['He', 'Li', 'Be', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'Cl', 'Ar', 'Ca', 'Sc', 'Ti',
                 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
                 'Rb', 'Sr', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn',
                 'Sb', 'Te', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
                 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'Re', 'Os', 'Ir', 'Pt',
                 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
                 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']

        if (label[0:2] in elem2):
            return label[0:2]

        elem1 = ['H', 'B', 'C', 'N', 'O', 'F', 'P', 'S', 'K', 'V', 'Y', 'I', 'W', 'U']

        if (label[0] in elem1):
            return label[0]

        print('WARNING: could not convert "%s" into element name!' % label)
        return label

    def get_symmetry_from_space_group_number(self, number):
        if number is not None:
            if number in [146, 148, 155, 160, 161, 166, 167]:
                if self.a != self.c:
                    return "HEXAGONAL"
                else:
                    return "RHOMBOHEDRAL"
            if number_between(number, 1, 2):
                return "TRICLINIC"
            if number_between(number, 3, 15):
                return "MONOCLINIC"
            if number_between(number, 16, 74):
                return "ORTHORHOMBIC"
            if number_between(number, 75, 142):
                return "TETRAGONAL"
            if number_between(number, 143, 167):
                return "TRIGONAL"
            if number_between(number, 168, 194):
                return "HEXAGONAL"
            if number_between(number, 195, 230):
                if self.alpha == 90 and self.beta == 90 and self.gamma == 90:
                    return "CUBIC"
                else:
                    return "RHOMBOHEDRAL"
        else:
            if self.alpha == 90 and self.beta == 90 and self.gamma == 90:
                if self.a == self.c and self.a == self.b:
                    return 'CUBIC'
                elif self.a == self.b:
                    return 'TETRAGONAL'
                else:
                    return 'ORTHORHOMBIC'
            elif self.alpha == 90 and self.beta == 90 and self.gamma == 120:
                if '6' in self.space_group:
                    return 'HEXAGONAL'
                elif '3' in self.space_group:
                    return 'TRIGONAL'
            elif self.alpha == 90 and self.gamma == 90:
                return 'MONOCLINIC'
            elif self.alpha == 90 and self.beta == 90:
                return 'MONOCLINIC'
        return 'TRICLINIC'

    def read_file_information(self):
        """
        Reads in all the header information and tries to build a good description of the phase.
        """
        if self.cif_dictionary.get('_chemical_formula_structural'):
            self.comments += self.cif_dictionary['_chemical_formula_structural'].replace(" ", "")
        elif self.cif_dictionary.get('_chemical_formula_analytical'):
            self.comments += self.cif_dictionary['_chemical_formula_analytical'].replace(" ", "")
        elif self.cif_dictionary.get('_chemical_formula_sum'):
            self.comments += self.cif_dictionary['_chemical_formula_sum'].replace(" ", "")

        if self.cif_dictionary.get('_symmetry_space_group_name_H-M'):
            if self.comments == '':
                self.comments += self.cif_dictionary.get('_symmetry_space_group_name_H-M').replace(" ", "")
            else:
                self.comments += ', ' + self.cif_dictionary.get('_symmetry_space_group_name_H-M').replace(" ", "")

        if self.cif_dictionary.get('_chemical_name_structure_type'):
            self.comments += ' - '
            self.comments += self.cif_dictionary['_chemical_name_structure_type']
            self.comments += ' structure type'
        if self.cif_dictionary.get('_database_code_icsd'):
            self.comments = self.comments.strip()
            if self.comments != '':
                self.comments += ', ICSD '
            else:
                self.comments += 'ICSD '
            self.comments += self.cif_dictionary['_database_code_icsd']
        elif self.cif_dictionary.get('_database_code_amcsd'):
            if self.comments != '':
                self.comments += ', amcsd '
            else:
                self.comments += 'amcsd '
            self.comments += self.cif_dictionary['_database_code_amcsd']


def number_between(number, low, high):
    """
    Tests if a number is in between low and high, whereby low and high are included  [low, high]
    :return: Boolean result for the result
    """
    if low <= number <= high:
        return True
    return False


def convert_cif_number_to_float(cif_number):
    return float(cif_number.split('(')[0])


def calculate_cell_volume(a, b, c, alpha, beta, gamma):
    """
    Calculates the cell volume using formula: 
    V = a*b*c*sqrt(1 + 2*cos(alpha)*cos(beta)*cos(gamma)-cos^2(alpha)-cos^2(beta)-cos^2(gamma))

    :param a: lattice parameter a
    :param b: lattice parameter b
    :param c: lattice parameter c
    :param alpha: lattice angle alpha in radians
    :param beta: lattice angle beta in radians
    :param gamma: lattice angle gamma in radians
    :return: cell volume
    """
    base = a * b * c
    part1 = np.cos(alpha) * np.cos(beta) * np.cos(gamma)
    part2 = np.cos(alpha) ** 2 + np.cos(beta) ** 2 + np.cos(gamma) ** 2
    return base * np.sqrt(1 + 2 * part1 - part2)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import h5py
import hdf5plugin


class Hdf5Image:
    def __init__(self, filename):
        """
        Loads an Hdf5 image produced by ESRF
        :param filename: path to the hdf5 file to be loaded
        """

        self.f = h5py.File(filename, 'r')
        self.image_sources = find_image_sources(self.f)

        self.dataset = self.f[self.image_sources[0]]
        self.series_max = self.dataset.shape[0]

    def get_image(self, ind):
        return self.dataset[ind]

    def select_source(self, source):
        self.dataset = self.f[source]
        self.series_max = self.dataset.shape[0]


def find_image_sources(hd5_file):
    image_paths = []

    def traverse_groups(group, parent_path=''):
        if isinstance(group, h5py.Dataset):
            if len(group.shape) >= 3:
                image_paths.append(parent_path)
        else:  # node is a group
            for key in group.keys():
                traverse_groups(group[key], parent_path + '/' + key)

    traverse_groups(hd5_file)

    return image_paths



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pyqtgraph as pg

from dioptas.controller.integration.overlay.OverlayInPatternController import OverlayInPatternController

from .BackgroundController import BackgroundController
from .CorrectionController import CorrectionController
from .ImageController import ImageController
from .overlay.OverlayController import OverlayController
from .PatternController import PatternController
from dioptas.controller.integration.phase.PhaseController import PhaseController
from .OptionsController import OptionsController
from .BatchController import BatchController

# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel

pg.setConfigOption('useOpenGL', False)
pg.setConfigOption('leftButtonPan', False)
pg.setConfigOption('background', 'k')
pg.setConfigOption('foreground', 'w')
pg.setConfigOption('antialias', True)


class IntegrationController(object):
    """
    This controller hosts all the Subcontroller of the integration tab.
    """

    def __init__(self, widget: IntegrationWidget, dioptas_model: DioptasModel):
        """
        :param widget: Reference to an IntegrationWidget
        :param dioptas_model: Reference to a DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """
        self.widget = widget
        self.model = dioptas_model

        self.create_sub_controller()

    def create_sub_controller(self):
        """
        Creates the sub controller with the appropriate data.
        """
        self.pattern_controller = PatternController(self.widget, self.model)
        self.image_controller = ImageController(self.widget, self.model)
        self.overlay_controller = OverlayController(self.widget, self.model)
        self.overlay_in_pattern_controller = OverlayInPatternController(self.widget.pattern_widget, self.model.overlay_model)
        self.phase_controller = PhaseController(self.widget, self.model)
        self.background_controller = BackgroundController(self.widget, self.model)
        self.correction_controller = CorrectionController(self.widget, self.model)
        self.options_controller = OptionsController(self.widget, self.model)
        self.batch_controller = BatchController(self.widget, self.model)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtCore, QtGui
import numpy as np

from ...widgets.CustomWidgets import NumberTextField, LabelAlignRight, DoubleSpinBoxAlignRight, HorizontalSpacerItem, \
    VerticalSpacerItem, CleanLooksComboBox

from ...model.util.HelperModule import convert_d_to_two_theta


class JcpdsEditorWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(JcpdsEditorWidget, self).__init__(parent)

        self.setWindowTitle('Dioptas - JCPDS Editor')

        self._layout = QtWidgets.QVBoxLayout()

        self._file_layout = QtWidgets.QGridLayout()
        self._file_layout.addWidget(LabelAlignRight('Filename:'), 0, 0)
        self._file_layout.addWidget(LabelAlignRight('Comment:'), 1, 0)

        self.filename_txt = QtWidgets.QLineEdit('')
        self.comments_txt = QtWidgets.QLineEdit('')
        self._file_layout.addWidget(self.filename_txt, 0, 1)
        self._file_layout.addWidget(self.comments_txt, 1, 1)
        self._layout.addLayout((self._file_layout))

        self.lattice_parameters_gb = QtWidgets.QGroupBox('Lattice Parameters')
        self._lattice_parameters_layout = QtWidgets.QVBoxLayout()

        self._symmetry_layout = QtWidgets.QHBoxLayout()
        self._symmetry_layout.addWidget(LabelAlignRight('Symmetry'))
        self.symmetry_cb = CleanLooksComboBox()
        self.symmetries = ['cubic', 'tetragonal', 'hexagonal', 'trigonal', 'rhombohedral',
                           'orthorhombic', 'monoclinic', 'triclinic']
        self.symmetry_cb.addItems(self.symmetries)
        self._symmetry_layout.addWidget(self.symmetry_cb)
        self._symmetry_layout.addSpacerItem(HorizontalSpacerItem())
        self._lattice_parameters_layout.addLayout(self._symmetry_layout)

        self._parameters_layout = QtWidgets.QGridLayout()

        self.lattice_a_sb = DoubleSpinBoxAlignRight()
        self.lattice_a_sb.setSingleStep(0.01)
        self.lattice_a_sb.setMinimum(0)
        self.lattice_a_sb.setMaximum(99999)
        self.lattice_a_sb.setDecimals(4)
        self.lattice_b_sb = DoubleSpinBoxAlignRight()
        self.lattice_b_sb.setMinimum(0)
        self.lattice_b_sb.setMaximum(99999)
        self.lattice_b_sb.setDecimals(4)
        self.lattice_b_sb.setSingleStep(0.01)
        self.lattice_c_sb = DoubleSpinBoxAlignRight()
        self.lattice_c_sb.setMinimum(0)
        self.lattice_c_sb.setMaximum(99999)
        self.lattice_c_sb.setDecimals(4)
        self.lattice_c_sb.setSingleStep(0.01)
        self.lattice_length_step_txt = NumberTextField('0.01')

        self.add_field(self._parameters_layout, self.lattice_a_sb, 'a0:', u"Å", 0, 0)
        self.add_field(self._parameters_layout, self.lattice_b_sb, 'b0:', u"Å", 0, 3)
        self.add_field(self._parameters_layout, self.lattice_c_sb, 'c0:', u"Å", 0, 6)
        self.add_field(self._parameters_layout, self.lattice_length_step_txt, 'st:', u"Å", 0, 9)

        self.lattice_eos_a_txt = NumberTextField()
        self.lattice_eos_b_txt = NumberTextField()
        self.lattice_eos_c_txt = NumberTextField()

        self.add_field(self._parameters_layout, self.lattice_eos_a_txt, 'a:', u"Å", 1, 0)
        self.add_field(self._parameters_layout, self.lattice_eos_b_txt, 'b:', u"Å", 1, 3)
        self.add_field(self._parameters_layout, self.lattice_eos_c_txt, 'c:', u"Å", 1, 6)

        self.lattice_alpha_sb = DoubleSpinBoxAlignRight()
        self.lattice_alpha_sb.setMaximum(180)
        self.lattice_beta_sb = DoubleSpinBoxAlignRight()
        self.lattice_beta_sb.setMaximum(180)
        self.lattice_gamma_sb = DoubleSpinBoxAlignRight()
        self.lattice_gamma_sb.setMaximum(180)
        self.lattice_angle_step_txt = NumberTextField('1')

        self.add_field(self._parameters_layout, self.lattice_alpha_sb, u'α:', u"°", 2, 0)
        self.add_field(self._parameters_layout, self.lattice_beta_sb, u'β:', u"°", 2, 3)
        self.add_field(self._parameters_layout, self.lattice_gamma_sb, u'γ:', u"°", 2, 6)
        self.add_field(self._parameters_layout, self.lattice_angle_step_txt, u'st:', u"°", 2, 9)

        self.lattice_volume_txt = NumberTextField()
        self.lattice_eos_volume_txt = NumberTextField()

        self.add_field(self._parameters_layout, self.lattice_volume_txt, 'V0:', u'Å³', 3, 3)
        self.add_field(self._parameters_layout, self.lattice_eos_volume_txt, 'V:', u'Å³', 3, 6)

        self.lattice_ab_sb = DoubleSpinBoxAlignRight()
        self.lattice_ab_sb.setDecimals(4)
        self.lattice_ca_sb = DoubleSpinBoxAlignRight()
        self.lattice_ca_sb.setDecimals(4)
        self.lattice_cb_sb = DoubleSpinBoxAlignRight()
        self.lattice_cb_sb.setDecimals(4)
        self.lattice_ratio_step_txt = NumberTextField('0.01')

        self.add_field(self._parameters_layout, self.lattice_ab_sb, 'a/b:', None, 4, 0)
        self.add_field(self._parameters_layout, self.lattice_ca_sb, 'c/a:', None, 4, 3)
        self.add_field(self._parameters_layout, self.lattice_cb_sb, 'c/b:', None, 4, 6)
        self.add_field(self._parameters_layout, self.lattice_ratio_step_txt, 'st:', None, 4, 9)

        self._lattice_parameters_layout.addLayout(self._parameters_layout)
        self.lattice_parameters_gb.setLayout(self._lattice_parameters_layout)

        self.eos_gb = QtWidgets.QGroupBox('Equation of State')
        self._eos_layout = QtWidgets.QGridLayout()

        self.eos_K_txt = NumberTextField()
        self.eos_Kp_txt = NumberTextField()
        self.eos_alphaT_txt = NumberTextField()
        self.eos_dalphadT_txt = NumberTextField()
        self.eos_dKdT_txt = NumberTextField()
        self.eos_dKpdT_txt = NumberTextField()

        self.add_field(self._eos_layout, self.eos_K_txt, 'K:', 'GPa', 0, 0)
        self.add_field(self._eos_layout, self.eos_Kp_txt, 'Kp:', None, 1, 0)
        self.add_field(self._eos_layout, self.eos_alphaT_txt, u'α<sub>T</sub>:', '1/K', 2, 0)
        self.add_field(self._eos_layout, self.eos_dalphadT_txt, u'dα<sub>T</sub>/dT:', u'1/K²', 3, 0)
        self.add_field(self._eos_layout, self.eos_dKdT_txt, 'dK/dT:', 'GPa/K', 4, 0)
        self.add_field(self._eos_layout, self.eos_dKpdT_txt, "dK'/dT", '1/K', 5, 0)
        self.eos_gb.setLayout(self._eos_layout)

        self.reflections_gb = QtWidgets.QGroupBox('Reflections')
        self._reflection_layout = QtWidgets.QGridLayout()
        self.reflection_table_view = QtWidgets.QTableView()
        self.reflection_table_model = ReflectionTableModel()
        self.reflection_table_view.setModel(self.reflection_table_model)
        # self.reflection_table.setColumnCount(10)
        self.reflections_add_btn = QtWidgets.QPushButton('Add')
        self.reflections_delete_btn = QtWidgets.QPushButton('Delete')
        self.reflections_clear_btn = QtWidgets.QPushButton('Clear')

        self._reflection_layout.addWidget(self.reflection_table_view, 0, 0, 1, 3)
        self._reflection_layout.addWidget(self.reflections_add_btn, 1, 0)
        self._reflection_layout.addWidget(self.reflections_delete_btn, 1, 1)
        self._reflection_layout.addWidget(self.reflections_clear_btn, 1, 2)

        self.reflections_gb.setLayout(self._reflection_layout)

        self._body_layout = QtWidgets.QGridLayout()
        self._body_layout.addWidget(self.eos_gb, 0, 0)
        self._body_layout.addItem(VerticalSpacerItem(), 1, 0)
        self._body_layout.addWidget(self.reflections_gb, 0, 1, 2, 1)

        self._button_layout = QtWidgets.QHBoxLayout()
        self.save_as_btn = QtWidgets.QPushButton('Save As')
        self.reload_file_btn = QtWidgets.QPushButton('Reload File')

        self._button_layout.addWidget(self.save_as_btn)
        self._button_layout.addWidget(self.reload_file_btn)
        self._button_layout.addSpacerItem(HorizontalSpacerItem())

        self._layout.addWidget(self.lattice_parameters_gb)
        self._layout.addLayout(self._body_layout)
        self._layout.addLayout(self._button_layout)
        self.setLayout(self._layout)

        self.style_widgets()

    def style_widgets(self):
        self.lattice_angle_step_txt.setMaximumWidth(60)
        self.lattice_length_step_txt.setMaximumWidth(60)
        self.lattice_ratio_step_txt.setMaximumWidth(60)

        self.reflection_table_view.setShowGrid(False)
        self.reflection_table_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.reflection_table_view.setItemDelegate(TextDoubleDelegate())
        self.reflection_table_view.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.eos_gb.setMaximumWidth(200)
        self.eos_gb.setStyleSheet("""
            QLineEdit {
                max-width: 80;
            }
        """)

        self.reflection_table_view.verticalHeader().setDefaultSectionSize(20)
        self.reflection_table_view.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        self.setWindowFlags(QtCore.Qt.Tool)
        self.setAttribute(QtCore.Qt.WA_MacAlwaysShowToolWindow)

    def raise_widget(self):
        self.show()
        self.setWindowState(self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        self.activateWindow()
        self.raise_()

    def add_field(self, layout, widget, label_str, unit, x, y):
        layout.addWidget(LabelAlignRight(label_str), x, y)
        layout.addWidget(widget, x, y + 1)
        if unit:
            layout.addWidget(QtWidgets.QLabel(unit), x, y + 2)

    def show_jcpds(self, jcpds_phase, wavelength=None):
        self.update_name(jcpds_phase)
        self.update_lattice_parameters(jcpds_phase)
        self.update_eos_parameters(jcpds_phase)
        self.reflection_table_model.update_reflection_data(jcpds_phase.reflections,
                                                           wavelength)

    def update_eos_parameters(self, jcpds_phase):
        self.eos_K_txt.setText(str(jcpds_phase.params['k0']))
        self.eos_Kp_txt.setText(str(jcpds_phase.params['k0p']))
        self.eos_alphaT_txt.setText(str(jcpds_phase.params['alpha_t0']))
        self.eos_dalphadT_txt.setText(str(jcpds_phase.params['d_alpha_dt']))
        self.eos_dKdT_txt.setText(str(jcpds_phase.params['dk0dt']))
        self.eos_dKpdT_txt.setText(str(jcpds_phase.params['dk0pdt']))

    def update_name(self, jcpds_phase):
        self.filename_txt.setText(jcpds_phase.filename)
        self.comments_txt.setText("/n".join(jcpds_phase.params['comments']))

    def update_lattice_parameters(self, jcpds_phase):
        self.blockAllSignals(True)
        self.symmetry_cb.setCurrentIndex(self.symmetries.index(jcpds_phase.params['symmetry'].lower()))
        self.update_spinbox_enable(jcpds_phase.params['symmetry'])

        if not self.lattice_a_sb.hasFocus():
            self.lattice_a_sb.setValue(jcpds_phase.params['a0'])
        if not self.lattice_b_sb.hasFocus():
            self.lattice_b_sb.setValue(jcpds_phase.params['b0'])
        if not self.lattice_c_sb.hasFocus():
            self.lattice_c_sb.setValue(jcpds_phase.params['c0'])

        self.lattice_eos_a_txt.setText('{0:.4f}'.format(jcpds_phase.params['a']))
        self.lattice_eos_b_txt.setText('{0:.4f}'.format(jcpds_phase.params['b']))
        self.lattice_eos_c_txt.setText('{0:.4f}'.format(jcpds_phase.params['c']))

        self.lattice_eos_volume_txt.setText('{0:.4f}'.format(jcpds_phase.params['v']))

        try:
            if not self.lattice_ab_sb.hasFocus():
                self.lattice_ab_sb.setValue(jcpds_phase.params['a0'] / float(jcpds_phase.params['b0']))
        except ZeroDivisionError:
            self.lattice_ab_sb.setSpecialValueText('Inf')

        try:
            if not self.lattice_ca_sb.hasFocus():
                self.lattice_ca_sb.setValue(jcpds_phase.params['c0'] / float(jcpds_phase.params['a0']))
        except ZeroDivisionError:
            self.lattice_ca_sb.setSpecialValueText('Inf')

        try:
            if not self.lattice_cb_sb.hasFocus():
                self.lattice_cb_sb.setValue(jcpds_phase.params['c0'] / float(jcpds_phase.params['b0']))
        except ZeroDivisionError:
            self.lattice_cb_sb.setSpecialValueText('Inf')

        self.lattice_volume_txt.setText(str('{0:g}'.format(jcpds_phase.params['v0'])))

        if not self.lattice_alpha_sb.hasFocus():
            self.lattice_alpha_sb.setValue(jcpds_phase.params['alpha0'])
        if not self.lattice_beta_sb.hasFocus():
            self.lattice_beta_sb.setValue(jcpds_phase.params['beta0'])
        if not self.lattice_gamma_sb.hasFocus():
            self.lattice_gamma_sb.setValue(jcpds_phase.params['gamma0'])

        self.blockAllSignals(False)

    def blockAllSignals(self, bool=True):
        self.lattice_a_sb.blockSignals(bool)
        self.lattice_b_sb.blockSignals(bool)
        self.lattice_c_sb.blockSignals(bool)

        self.lattice_alpha_sb.blockSignals(bool)
        self.lattice_beta_sb.blockSignals(bool)
        self.lattice_gamma_sb.blockSignals(bool)

        self.lattice_ab_sb.blockSignals(bool)
        self.lattice_ca_sb.blockSignals(bool)
        self.lattice_cb_sb.blockSignals(bool)

        self.symmetry_cb.blockSignals(bool)

    def update_spinbox_enable(self, symmetry):
        if symmetry == 'CUBIC':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(False)
            self.lattice_c_sb.setEnabled(False)

            self.lattice_alpha_sb.setEnabled(False)
            self.lattice_beta_sb.setEnabled(False)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(False)
            self.lattice_ca_sb.setEnabled(False)
            self.lattice_cb_sb.setEnabled(False)

        elif symmetry == 'TETRAGONAL':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(False)
            self.lattice_c_sb.setEnabled(True)

            self.lattice_alpha_sb.setEnabled(False)
            self.lattice_beta_sb.setEnabled(False)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(False)
            self.lattice_ca_sb.setEnabled(True)
            self.lattice_cb_sb.setEnabled(False)

        elif symmetry == 'ORTHORHOMBIC':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(True)
            self.lattice_c_sb.setEnabled(True)

            self.lattice_alpha_sb.setEnabled(False)
            self.lattice_beta_sb.setEnabled(False)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(True)
            self.lattice_ca_sb.setEnabled(True)
            self.lattice_cb_sb.setEnabled(True)

        elif symmetry == 'HEXAGONAL' or symmetry == 'TRIGONAL':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(False)
            self.lattice_c_sb.setEnabled(True)

            self.lattice_alpha_sb.setEnabled(False)
            self.lattice_beta_sb.setEnabled(False)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(False)
            self.lattice_ca_sb.setEnabled(True)
            self.lattice_cb_sb.setEnabled(False)

        elif symmetry == 'RHOMBOHEDRAL':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(False)
            self.lattice_c_sb.setEnabled(False)

            self.lattice_alpha_sb.setEnabled(True)
            self.lattice_beta_sb.setEnabled(False)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(False)
            self.lattice_ca_sb.setEnabled(False)
            self.lattice_cb_sb.setEnabled(False)

        elif symmetry == 'MONOCLINIC':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(True)
            self.lattice_c_sb.setEnabled(True)

            self.lattice_alpha_sb.setEnabled(False)
            self.lattice_beta_sb.setEnabled(True)
            self.lattice_gamma_sb.setEnabled(False)

            self.lattice_ab_sb.setEnabled(True)
            self.lattice_ca_sb.setEnabled(True)
            self.lattice_cb_sb.setEnabled(True)

        elif symmetry == 'TRICLINIC':
            self.lattice_a_sb.setEnabled(True)
            self.lattice_b_sb.setEnabled(True)
            self.lattice_c_sb.setEnabled(True)

            self.lattice_alpha_sb.setEnabled(True)
            self.lattice_beta_sb.setEnabled(True)
            self.lattice_gamma_sb.setEnabled(True)

            self.lattice_ab_sb.setEnabled(True)
            self.lattice_ca_sb.setEnabled(True)
            self.lattice_cb_sb.setEnabled(True)

        else:
            print('Unknown symmetry: {0}.'.format(symmetry))

    def get_selected_reflections(self):
        selected = self.reflection_table_view.selectionModel().selectedRows()
        try:
            row = []
            for element in selected:
                row.append(int(element.row()))
        except IndexError:
            row = None
        return row


class NoRectDelegate(QtWidgets.QItemDelegate):
    def drawFocus(self, painter, option, rect):
        option.state &= ~QtWidgets.QStyle.State_HasFocus
        QtWidgets.QItemDelegate.drawFocus(self, painter, option, rect)


class TextDoubleDelegate(NoRectDelegate):
    def createEditor(self, parent, _, model):
        self.editor = QtWidgets.QLineEdit(parent)
        self.editor.setFrame(False)
        self.editor.setValidator(QtGui.QDoubleValidator())
        self.editor.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        return self.editor


class ReflectionTableModel(QtCore.QAbstractTableModel):
    reflection_edited = QtCore.Signal(int, int, str)  # row, column, value

    def __init__(self, reflections=None, wavelength=None):
        super(ReflectionTableModel, self).__init__()
        self.wavelength = wavelength
        if reflections is not None:
            self.reflections = reflections
            self.update_reflection_data(reflections)
        else:
            self.reflections = []
        self.header_labels = ['h', 'k', 'l', 'Intensity', 'd0', 'd', u"2θ_0", u"2θ", 'Q0', 'Q']

    def rowCount(self, *_):
        return len(self.reflections)

    def columnCount(self, *_):
        return 10

    def data(self, index, role=QtCore.Qt.DisplayRole):
        col = index.column()
        if role == QtCore.Qt.TextAlignmentRole:
            return QtCore.Qt.AlignCenter
        if role == QtCore.Qt.DisplayRole:
            if col < 4:
                format_str = '{0:g}'
            else:
                format_str = '{0:.4f}'
            return format_str.format(self.reflection_data[index.row(), index.column()])
        else:
            return QtCore.QVariant()

    def setData(self, index, value, role):
        self.reflection_edited.emit(index.row(), index.column(), value)
        return True

    def update_reflection_data(self, reflections, wavelength=None):
        if wavelength is None:
            wavelength = self.wavelength
        else:
            self.wavelength = wavelength

        cur_row_num = self.rowCount()
        row_diff = len(reflections) - cur_row_num
        if row_diff < 0:
            self.beginRemoveRows(QtCore.QModelIndex(), cur_row_num + row_diff, cur_row_num - 1)
        elif row_diff > 0:
            self.beginInsertRows(QtCore.QModelIndex(), cur_row_num, cur_row_num + row_diff - 1)

        self.reflections = reflections
        self.reflection_data = np.zeros((len(reflections), self.columnCount()))
        for i, refl in enumerate(reflections):
            self.reflection_data[i, 0] = refl.h
            self.reflection_data[i, 1] = refl.k
            self.reflection_data[i, 2] = refl.l
            self.reflection_data[i, 3] = refl.intensity
            self.reflection_data[i, 4] = refl.d0
            self.reflection_data[i, 5] = refl.d

        if wavelength is not None:
            self.reflection_data[:, 6] = convert_d_to_two_theta(self.reflection_data[:, 4], wavelength)  # two_theta0
            self.reflection_data[:, 7] = convert_d_to_two_theta(self.reflection_data[:, 5], wavelength)  # two_theta
            valid_ind = np.where(self.reflection_data[:, 4] > 0)
            self.reflection_data[valid_ind, 8] = 2.0 * np.pi / self.reflection_data[valid_ind, 4]  # q0
            self.reflection_data[valid_ind, 9] = 2.0 * np.pi / self.reflection_data[valid_ind, 5]  # q

        if row_diff < 0:
            self.endRemoveRows()
        elif row_diff > 0:
            self.endInsertRows()

        self.modelReset.emit()

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int = ...):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self.header_labels[section]
        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return section + 1

    def flags(self, index):
        col = index.column()
        ans = QtCore.QAbstractTableModel.flags(self, index)
        if col <= 3:
            return QtCore.Qt.ItemIsEditable | ans
        else:
            return ans



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtGui, QtCore

from dioptas.widgets.integration import CLICKED_COLOR

from ..CustomWidgets import LabelAlignRight, FlatButton, CleanLooksComboBox


class MouseCurrentAndClickedWidget(QtWidgets.QWidget):
    def __init__(self, clicked_color):
        super(MouseCurrentAndClickedWidget, self).__init__()

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        self.cur_pos_widget = MousePositionWidget()
        self.clicked_pos_widget = MousePositionWidget(clicked_color)

        self._layout.addWidget(self.cur_pos_widget)
        self._layout.addWidget(self.clicked_pos_widget)

        self.setLayout(self._layout)


class MousePositionWidget(QtWidgets.QWidget):
    def __init__(self, color=None):
        super(MousePositionWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.x_pos_lbl = QtWidgets.QLabel("X:")
        self.y_pos_lbl = QtWidgets.QLabel("Y:")
        self.int_lbl = QtWidgets.QLabel("I:")

        self._layout.addWidget(self.x_pos_lbl)
        self._layout.addWidget(self.y_pos_lbl)
        self._layout.addWidget(self.int_lbl)

        self.setLayout(self._layout)
        self.style_widgets(color)

    def style_widgets(self, color):
        main_style_str = """
            QLabel {
                min-width: 70px;
            }
        """
        self.setStyleSheet(main_style_str)

        if color is not None:
            style_str = "color: {};".format(color)
            self.x_pos_lbl.setStyleSheet(style_str)
            self.y_pos_lbl.setStyleSheet(style_str)
            self.int_lbl.setStyleSheet(style_str)


class MouseUnitCurrentAndClickedWidget(QtWidgets.QWidget):
    def __init__(self, clicked_color=CLICKED_COLOR):
        super(MouseUnitCurrentAndClickedWidget, self).__init__()
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        self.cur_unit_widget = MouseUnitWidget()
        self.clicked_unit_widget = MouseUnitWidget(clicked_color)

        self._layout.addWidget(self.cur_unit_widget)
        self._layout.addWidget(self.clicked_unit_widget)

        self.setLayout(self._layout)


class MouseUnitWidget(QtWidgets.QWidget):
    def __init__(self, color=None):
        super(MouseUnitWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.tth_lbl = QtWidgets.QLabel("2θ:")
        self.q_lbl = QtWidgets.QLabel("Q:")
        self.d_lbl = QtWidgets.QLabel("d:")
        self.azi_lbl = QtWidgets.QLabel("X:")

        self._layout.addWidget(self.tth_lbl)
        self._layout.addWidget(self.q_lbl)
        self._layout.addWidget(self.d_lbl)
        self._layout.addWidget(self.azi_lbl)

        self.setLayout(self._layout)
        self.style_widgets(color)

    def style_widgets(self, color):
        main_style_str = """
            QLabel {
                min-width: 70px;
            }
        """
        self.setStyleSheet(main_style_str)

        if color is not None:
            style_str = "color: {};".format(color)
            self.tth_lbl.setStyleSheet(style_str)
            self.d_lbl.setStyleSheet(style_str)
            self.q_lbl.setStyleSheet(style_str)
            self.azi_lbl.setStyleSheet(style_str)


class BrowseFileWidget(QtWidgets.QWidget):
    def __init__(self, files, checkbox_text):
        super(BrowseFileWidget, self).__init__()

        self._layout = QtWidgets.QGridLayout()
        self._layout.setContentsMargins(0, 5, 5, 0)
        self._layout.setVerticalSpacing(3)
        self._layout.setHorizontalSpacing(3)

        self.load_btn = QtWidgets.QPushButton("Load {}(s)".format(files))
        self.file_cb = QtWidgets.QCheckBox(checkbox_text)

        self._load_layout = QtWidgets.QVBoxLayout()
        self._load_layout.setContentsMargins(0, 0, 0, 0)
        self._load_layout.setSpacing(3)
        self._load_layout.addWidget(self.load_btn)
        self._load_layout.addWidget(self.file_cb)

        self._layout.addLayout(self._load_layout, 0, 0, 2, 1)

        self.directory_txt = QtWidgets.QLineEdit("")
        self.directory_btn = QtWidgets.QPushButton("...")
        self.file_txt = QtWidgets.QLineEdit("")

        self.step_file_widget = StepFileWidget()
        self._layout.addWidget(self.step_file_widget, 0, 1, 2, 2)
        self.step_series_widget = StepFrameWidget()
        self._layout.addWidget(self.step_series_widget, 0, 3, 2, 2)

        self.step_series_widget.setVisible(False)

        self._layout.addWidget(self.file_txt, 2, 0, 1, 5)
        self._directory_layout = QtWidgets.QHBoxLayout()
        self._directory_layout.addWidget(self.directory_txt)
        self._directory_layout.addWidget(self.directory_btn)
        self._directory_layout.setSpacing(3)
        self._layout.addLayout(self._directory_layout, 3, 0, 1, 5)

        self.sources_widget = QtWidgets.QWidget()
        self.sources_cb = CleanLooksComboBox()
        self._sources_layout = QtWidgets.QHBoxLayout()
        self._sources_layout.setContentsMargins(0, 0, 0, 0)
        self._sources_layout.addWidget(LabelAlignRight("Source:"))
        self._sources_layout.addWidget(self.sources_cb)
        self.sources_widget.setLayout(self._sources_layout)
        self._layout.addWidget(self.sources_widget, 4, 0, 1, 5)

        self.setLayout(self._layout)

        self.style_widgets()

    def style_widgets(self):
        self.load_btn.setMaximumWidth(120)
        self.load_btn.setMinimumWidth(120)
        self.load_btn.setFixedHeight(25)

        small_btn_width = 25
        self.directory_btn.setFixedSize(small_btn_width, small_btn_width)
        self.sources_widget.hide()


class StepWidget(QtWidgets.QWidget):
    """Widget with buttons to step a number up and down and a Spinbox with the step size"""

    iteration_name = ""

    def __init__(self):
        super(StepWidget, self).__init__()
        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

        self._layout = QtWidgets.QGridLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(3)
        self.init_navigator()

        self.setLayout(self._layout)
        self.style_widgets()

    def init_navigator(self):
        self.next_btn = QtWidgets.QPushButton(">")
        self.next_btn.setToolTip("Loads next {}".format(self.iteration_name))
        self.previous_btn = QtWidgets.QPushButton("<")
        self.previous_btn.setToolTip(("Loads previous {}".format(self.iteration_name)))
        self.step_txt = QtWidgets.QSpinBox()
        self.step_txt.setValue(1)
        self.step_txt.setRange(1, 10000)
        self.step_txt.setToolTip("Step size")

        self._layout.addWidget(self.previous_btn, 0, 0)
        self._layout.addWidget(self.next_btn, 0, 1)
        self._step_layout = QtWidgets.QHBoxLayout()
        self._step_layout.addWidget(LabelAlignRight("Step:"))
        self._step_layout.addWidget(self.step_txt)
        self._layout.addLayout(self._step_layout, 1, 0, 1, 2)

    def style_widgets(self):
        self.step_txt.setMaximumWidth(53)
        self.step_txt.setFixedHeight(25)

        small_btn_width = 40
        small_btn_height = 25
        btns = [self.next_btn, self.previous_btn]
        for btn in btns:
            btn.setFixedHeight(small_btn_height)
            btn.setFixedWidth(small_btn_width)


class StepFileWidget(StepWidget):
    iteration_name = "file"

    def __init__(self):
        super(StepFileWidget, self).__init__()
        self.browse_by_name_rb = QtWidgets.QRadioButton("By Name")
        self.browse_by_name_rb.setChecked(True)
        self.browse_by_time_rb = QtWidgets.QRadioButton("By Time")

        self._layout.addWidget(self.browse_by_name_rb, 0, 2)
        self._layout.addWidget(self.browse_by_time_rb, 1, 2)

        self.set_tooltips()

    def set_tooltips(self):
        self.browse_by_name_rb.setToolTip("Browse files by name")
        self.browse_by_time_rb.setToolTip("Browse files by time")


class StepFrameWidget(StepWidget):
    iteration_name = "frame"

    def __init__(self):
        super(StepFrameWidget, self).__init__()
        self.pos_txt = QtWidgets.QLineEdit()
        self.pos_validator = QtGui.QIntValidator(1, 1)
        self.pos_txt.setValidator(self.pos_validator)
        self.pos_txt.setToolTip("Currently loaded frame")
        self.pos_label = QtWidgets.QLabel("Frame:")
        self.pos_label.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignBottom)

        self._pos_layout = QtWidgets.QVBoxLayout()
        self._pos_layout.addWidget(self.pos_label)
        self._pos_layout.addWidget(self.pos_txt)
        self._layout.addLayout(self._pos_layout, 0, 2, 2, 2)

        self.pos_txt.setSizePolicy(
            QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Maximum
        )



from qtpy import QtCore, QtGui, QtWidgets
import pyqtgraph as pq
from pyqtgraph import GraphicsLayoutWidget
import os
import numpy as np
from .plot_widgets.HistogramLUTItem import HistogramLUTItem

from .CustomWidgets import FlatButton
from .. import style_path


class Map2DWidget(QtWidgets.QWidget):
    widget_raised = QtCore.Signal()
    widget_closed = QtCore.Signal()

    mouse_clicked = QtCore.Signal(float, float)
    mouse_moved = QtCore.Signal(float, float)

    def __init__(self, parent=None):
        super(Map2DWidget, self).__init__(parent)
        # setup MAP window
        self.setWindowTitle("2D Map")
        # self.setGeometry(200, 100, 800, 600)

        self.create_main_control_widgets()
        self.create_main_control_layout()

        self.create_map_widgets()
        self.create_map_layout()

        self.create_background_control_widgets()
        self.create_background_control_layout()

        self.create_roi_widgets()
        self.create_roi_layout()
        self.hide_roi_controls()

        self.create_status_bar()
        self.create_status_layout()

        self.create_layout()
        self.create_tooltips()
        self.style_widgets()

        self.setWindowFlags(QtCore.Qt.Dialog | QtCore.Qt.WindowCloseButtonHint |
                            QtCore.Qt.WindowMaximizeButtonHint)

        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.MaximizeUsingFullscreenGeometryHint)
        self.setAttribute(QtCore.Qt.WA_MacAlwaysShowToolWindow)

        # Map mouse events
        self.map_image.mouseClickEvent = self.mouse_click_event
        self.map_view_box.mouseClickEvent = self.do_nothing
        self.hist_layout.scene().sigMouseMoved.connect(self.map_mouse_move_event)

    def create_main_control_widgets(self):
        self.load_map_files_btn = FlatButton("Load Map Files")
        self.setup_map_btn = FlatButton("Setup Map")

        self.auto_update_map_cb = QtWidgets.QCheckBox('Auto Update?')
        self.auto_update_map_cb.setChecked(True)
        self.update_map_btn = QtWidgets.QPushButton("Update Map")

        self.mode_gb = QtWidgets.QGroupBox("Mode")
        self.interactive_mode_rb = QtWidgets.QRadioButton('Interactive')
        self.interactive_mode_rb.setChecked(True)
        self.roi_mode_rb = QtWidgets.QRadioButton('ROI')

    def create_main_control_layout(self):
        self.main_control_layout = QtWidgets.QVBoxLayout()
        self.main_control_layout.addWidget(self.load_map_files_btn)
        self.main_control_layout.addWidget(self.setup_map_btn)

        self.update_layout = QtWidgets.QHBoxLayout()
        self.update_layout.addWidget(self.auto_update_map_cb)
        self.update_layout.addWidget(self.update_map_btn)
        self.main_control_layout.addLayout(self.update_layout)

        self.mode_gb_layout = QtWidgets.QHBoxLayout()
        self.mode_gb_layout.addWidget(self.interactive_mode_rb)
        self.mode_gb_layout.addWidget(self.roi_mode_rb)
        self.mode_gb.setLayout(self.mode_gb_layout)
        self.main_control_layout.addWidget(self.mode_gb)

    def create_map_widgets(self):
        self.map_image = pq.ImageItem()
        self.map_histogram_LUT = HistogramLUTItem(self.map_image, orientation='vertical')

        self.bg_image = np.zeros([1920, 1200])
        self.map_bg_image = pq.ImageItem()
        bg_rect = QtCore.QRectF(0, 0, 1920, 1200)
        self.map_bg_image.setImage(self.bg_image, opacity=0.5)
        self.map_bg_image.setRect(bg_rect)
        self.snapshot_btn = QtWidgets.QPushButton("Snapshot")

    def create_map_layout(self):
        self.hist_layout = GraphicsLayoutWidget()
        self.map_view_box = self.hist_layout.addViewBox(0, 0, lockAspect=1.0)

        self.map_view_box.addItem(self.map_bg_image, ignoreBounds=True)  # MAPBG
        self.map_view_box.addItem(self.map_image)
        # self.map_hor_axis = pq.AxisItem(orientation='bottom')
        # self.hist_layout.addItem(self.map_hor_axis)
        # self.map_ver_axis = pq.AxisItem(orientation='left')
        # self.hist_layout.addItem(self.map_ver_axis)
        self.map_histogram_LUT = HistogramLUTItem(self.map_image, orientation='vertical')
        self.hist_layout.addItem(self.map_histogram_LUT, 0, 1)

    def create_background_control_widgets(self):
        self.add_bg_btn = QtWidgets.QPushButton("Add BG Image")
        self.bg_opacity_lbl = QtWidgets.QLabel("Opacity: BG")
        self.bg_opacity_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.map_opacity_lbl = QtWidgets.QLabel("Map")

    def create_background_control_layout(self):
        self.background_layout = QtWidgets.QHBoxLayout()
        self.background_layout.addWidget(self.add_bg_btn)
        self.background_layout.addWidget(self.bg_opacity_lbl)
        self.background_layout.addWidget(self.bg_opacity_slider)
        self.background_layout.addWidget(self.map_opacity_lbl)
        self.background_layout.addStretch(1)

    def create_status_bar(self):
        self.map_filename_lbl = QtWidgets.QLabel()
        self.map_pos_lbl = QtWidgets.QLabel()

    def create_status_layout(self):
        self.status_layout = QtWidgets.QHBoxLayout()
        self.status_layout.addStretch(1)
        self.status_layout.addWidget(self.map_filename_lbl)
        self.status_layout.addWidget(self.map_pos_lbl)

    def create_roi_widgets(self):
        self.roi_widget = QtWidgets.QWidget()
        self.roi_list = QtWidgets.QListWidget()
        self.roi_list.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.roi_math_lbl = QtWidgets.QLabel('Math:')
        self.roi_math_txt = QtWidgets.QLineEdit()
        self.roi_add_btn = QtWidgets.QPushButton("Add Range")
        self.roi_add_phase_btn = QtWidgets.QPushButton("Add Phase")
        self.roi_del_btn = QtWidgets.QPushButton("Remove Range(s)")
        self.roi_clear_btn = QtWidgets.QPushButton("Clear Ranges")
        self.roi_toggle_btn = QtWidgets.QPushButton("Toggle Show")
        self.roi_select_all_btn = QtWidgets.QPushButton("Select All")

    def create_roi_layout(self):
        self.roi_vbox_layout = QtWidgets.QVBoxLayout()
        self.roi_vbox_layout.addWidget(self.roi_list)
        self.roi_grid_layout = QtWidgets.QGridLayout()
        self.roi_grid_layout.addWidget(self.roi_add_btn, 0, 0)
        self.roi_grid_layout.addWidget(self.roi_add_phase_btn, 0, 1)
        self.roi_grid_layout.addWidget(self.roi_del_btn, 1, 0)
        self.roi_grid_layout.addWidget(self.roi_clear_btn, 1, 1)
        self.roi_grid_layout.addWidget(self.roi_toggle_btn, 2, 1)
        self.roi_grid_layout.addWidget(self.roi_select_all_btn, 2, 0)

        self.math_layout = QtWidgets.QHBoxLayout()
        self.math_layout.addWidget(self.roi_math_lbl)
        self.math_layout.addWidget(self.roi_math_txt)
        self.roi_vbox_layout.addLayout(self.math_layout)
        self.roi_vbox_layout.addLayout(self.roi_grid_layout)

        self.roi_widget.setLayout(self.roi_vbox_layout)

        self.main_control_layout.addWidget(self.roi_widget)
        self.main_control_layout.addStretch(1)

    def create_layout(self):
        self.main_vertical_layout = QtWidgets.QVBoxLayout()
        self.main_horizontal_layout = QtWidgets.QHBoxLayout()
        self.main_horizontal_layout.addLayout(self.main_control_layout)
        self.main_horizontal_layout.addWidget(self.hist_layout)

        self.main_vertical_layout.addLayout(self.main_horizontal_layout)
        self.main_vertical_layout.addLayout(self.status_layout)

        # self.main_vbox.addStretch(1)
        # self.main_vbox.addLayout(self.status_hbox)
        self.setLayout(self.main_vertical_layout)

    def create_tooltips(self):
        self.load_map_files_btn.setToolTip("Load the image or pattern files for your map")
        self.setup_map_btn.setToolTip('Setup the Coordinates of the Map manually')

    def style_widgets(self):
        # Widget Properties
        self.roi_toggle_btn.setCheckable(True)
        self.roi_toggle_btn.setChecked(True)
        self.bg_opacity_slider.setMinimum(0)
        self.bg_opacity_slider.setMaximum(100)
        self.bg_opacity_slider.setValue(50)
        self.bg_opacity_slider.setSingleStep(5)
        self.bg_opacity_slider.setPageStep(20)

    def raise_widget(self):
        self.show()
        self.setWindowState(self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        self.activateWindow()
        self.raise_()
        self.widget_raised.emit()

    def show_roi_controls(self):
        self.roi_widget.show()

    def hide_roi_controls(self):
        self.roi_widget.hide()

    def closeEvent(self, a0: QtGui.QCloseEvent):
        self.widget_closed.emit()
        super(Map2DWidget, self).closeEvent(a0)

    def mouse_click_event(self, ev):
        if ev.button() == QtCore.Qt.RightButton or \
                (ev.button() == QtCore.Qt.LeftButton and ev.modifiers() & QtCore.Qt.ControlModifier):
            self.map_view_box.autoRange()

        elif ev.button() == QtCore.Qt.LeftButton:
            pos = ev.pos()
            x = pos.x()
            y = pos.y()
            self.mouse_clicked.emit(x, y)

    def map_mouse_move_event(self, pos):
        pos = self.map_image.mapFromScene(pos)
        x = pos.x()
        y = pos.y()
        self.mouse_moved.emit(x, y)

    # prevents right-click from opening menu
    def do_nothing(self, ev):
        pass


class SetupMapDialog(QtWidgets.QDialog):
    """
    Dialog for inputting map positions manually
    """

    def __init__(self, parent):
        super(SetupMapDialog, self).__init__(parent)

        self.setWindowTitle("Setup Map")

        self._parent = parent
        self._create_widgets()
        self._layout_widgets()
        self._style_widgets()

        self.setWindowFlags(QtCore.Qt.Tool)
        self.setAttribute(QtCore.Qt.WA_MacAlwaysShowToolWindow)

        self._connect_widgets()
        self.approved = False

    def _create_widgets(self):
        self.selected_map_files = QtWidgets.QListWidget()
        self.read_list_btn = QtWidgets.QPushButton('Read Files')
        self.move_up_btn = QtWidgets.QPushButton(u'\u2191')
        self.move_down_btn = QtWidgets.QPushButton(u'\u2193')
        self.add_empty_btn = QtWidgets.QPushButton('Empty')
        self.delete_btn = QtWidgets.QPushButton('Delete')

        self.hor_lbl = QtWidgets.QLabel("Horizontal")
        self.ver_lbl = QtWidgets.QLabel("Vertical")
        self.min_lbl = QtWidgets.QLabel("Minimum")
        self.step_lbl = QtWidgets.QLabel("Step")
        self.num_lbl = QtWidgets.QLabel("Number")
        self.first_lbl = QtWidgets.QLabel("First")
        self.total_files_lbl = QtWidgets.QLabel("0 Files")
        self.total_map_points_lbl = QtWidgets.QLabel("0 points")

        self.hor_min_txt = QtWidgets.QLineEdit()
        self.hor_step_txt = QtWidgets.QLineEdit()
        self.hor_num_txt = QtWidgets.QLineEdit()
        self.ver_min_txt = QtWidgets.QLineEdit()
        self.ver_step_txt = QtWidgets.QLineEdit()
        self.ver_num_txt = QtWidgets.QLineEdit()
        self.first_hor_rb = QtWidgets.QRadioButton("Horizontal")
        self.first_ver_rb = QtWidgets.QRadioButton("Vertical")

        self.first_hor_rb.setToolTip("Horizontal position changes between first two files")
        self.first_ver_rb.setToolTip("Vertical position changes between first two files")

        self.min_unit_lbl = QtWidgets.QLabel("mm")
        self.step_unit_lbl = QtWidgets.QLabel("mm")

        self.ok_btn = FlatButton("Done")
        self.cancel_btn = FlatButton("Cancel")

    def _layout_widgets(self):
        self._grid_layout = QtWidgets.QGridLayout()
        self._hbox_layout = QtWidgets.QHBoxLayout()
        self._vbox_list_layout = QtWidgets.QVBoxLayout()
        self._vbox_list_controls_layout = QtWidgets.QVBoxLayout()

        self._grid_layout.addWidget(self.hor_lbl, 0, 1)
        self._grid_layout.addWidget(self.ver_lbl, 0, 2)
        self._grid_layout.addWidget(self.min_lbl, 1, 0)
        self._grid_layout.addWidget(self.hor_min_txt, 1, 1)
        self._grid_layout.addWidget(self.ver_min_txt, 1, 2)
        self._grid_layout.addWidget(self.min_unit_lbl, 1, 3)
        self._grid_layout.addWidget(self.step_lbl, 2, 0)
        self._grid_layout.addWidget(self.hor_step_txt, 2, 1)
        self._grid_layout.addWidget(self.ver_step_txt, 2, 2)
        self._grid_layout.addWidget(self.step_unit_lbl, 2, 3)
        self._grid_layout.addWidget(self.num_lbl, 3, 0)
        self._grid_layout.addWidget(self.hor_num_txt, 3, 1)
        self._grid_layout.addWidget(self.ver_num_txt, 3, 2)
        self._grid_layout.addWidget(self.total_map_points_lbl, 3, 3)
        self._grid_layout.addWidget(self.first_lbl, 4, 0)
        self._grid_layout.addWidget(self.first_hor_rb, 4, 1)
        self._grid_layout.addWidget(self.first_ver_rb, 4, 2)
        self._grid_layout.addWidget(self.ok_btn, 5, 2)
        self._grid_layout.addWidget(self.cancel_btn, 5, 3)

        self._hbox_layout.addLayout(self._grid_layout)

        self._vbox_list_layout.addWidget(self.selected_map_files)
        self._vbox_list_layout.addWidget(self.total_files_lbl)
        self._hbox_layout.addLayout(self._vbox_list_layout)

        self._vbox_list_controls_layout.addWidget(self.read_list_btn)
        self._vbox_list_controls_layout.addWidget(self.move_up_btn)
        self._vbox_list_controls_layout.addWidget(self.add_empty_btn)
        self._vbox_list_controls_layout.addWidget(self.delete_btn)
        self._vbox_list_controls_layout.addWidget(self.move_down_btn)

        self._hbox_layout.addLayout(self._vbox_list_controls_layout)
        self.setLayout(self._hbox_layout)

    def _style_widgets(self):
        """
        Makes everything pretty and set double/int validators for the line edits.
        """

        self.selected_map_files.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        self.hor_min_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.hor_min_txt.setMaximumWidth(40)
        self.ver_min_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.ver_min_txt.setMaximumWidth(40)
        self.hor_step_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.hor_step_txt.setMaximumWidth(40)
        self.ver_step_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.ver_step_txt.setMaximumWidth(40)
        self.hor_num_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.hor_num_txt.setMaximumWidth(40)
        self.ver_num_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.ver_num_txt.setMaximumWidth(40)

        self.hor_min_txt.setValidator(QtGui.QDoubleValidator())
        self.ver_min_txt.setValidator(QtGui.QDoubleValidator())
        self.hor_step_txt.setValidator(QtGui.QDoubleValidator())
        self.ver_step_txt.setValidator(QtGui.QDoubleValidator())
        self.hor_num_txt.setValidator(QtGui.QIntValidator())
        self.ver_num_txt.setValidator(QtGui.QIntValidator())

        self.ok_btn.setEnabled(False)

    def _connect_widgets(self):
        """
        Connecting actions to slots.
        """
        self.ok_btn.clicked.connect(self.accept_manual_map_positions)
        self.cancel_btn.clicked.connect(self.reject_manual_map_positions)

    def accept_manual_map_positions(self):
        self.approved = True
        self.accept()

    def reject_manual_map_positions(self):
        self.approved = False
        self.reject()

    @property
    def hor_minimum(self):
        return float(str(self.hor_min_txt.text()))

    @property
    def ver_minimum(self):
        return float(str(self.ver_min_txt.text()))

    @property
    def hor_step_size(self):
        return float(str(self.hor_step_txt.text()))

    @property
    def ver_step_size(self):
        return float(str(self.ver_step_txt.text()))

    @property
    def hor_number(self):
        return int(str(self.hor_num_txt.text()))

    @property
    def ver_number(self):
        return int(str(self.ver_num_txt.text()))

    @property
    def is_hor_first(self):
        """
        Returns: True if horizontal is first, False if vertical is first
        """
        return self.first_hor_rb.isChecked()

    def exec_(self):
        """
        Overwriting the dialog exec_ function to center the widget in the parent window before execution.
        """
        parent_center = self._parent.window().mapToGlobal(self._parent.window().rect().center())
        self.move(parent_center.x() - 101, parent_center.y() - 48)
        super(SetupMapDialog, self).exec_()


class OpenBGImageDialog(QtWidgets.QDialog):
    """
    Dialog for setting up a background image
    """

    def __init__(self, parent, default_config):
        super(OpenBGImageDialog, self).__init__()

        self._parent = parent
        self._default_config = default_config
        self._create_widgets()
        self._layout_widgets()
        self._style_widgets()

        self._connect_widgets()
        self.approved = False

    def _create_widgets(self):
        self.bg_file_name_lbl = QtWidgets.QLabel()
        self.hor_lbl = QtWidgets.QLabel("Horizontal")
        self.ver_lbl = QtWidgets.QLabel("Vertical")
        self.img_px_lbl = QtWidgets.QLabel("Number of Pixels")
        self.img_px_size_lbl = QtWidgets.QLabel("Pixel Size")
        self.img_center_lbl = QtWidgets.QLabel("Center Position")
        self.flip_lbl = QtWidgets.QLabel("Flip?")

        self.img_hor_px_txt = QtWidgets.QLineEdit(str(self._default_config['img_hor_px']))
        self.img_ver_px_txt = QtWidgets.QLineEdit(str(self._default_config['img_ver_px']))
        self.img_hor_px_size_txt = QtWidgets.QLineEdit(str(self._default_config['img_px_size_hor']))
        self.img_ver_px_size_txt = QtWidgets.QLineEdit(str(self._default_config['img_px_size_ver']))
        self.img_hor_center_txt = QtWidgets.QLineEdit()
        self.img_ver_center_txt = QtWidgets.QLineEdit()
        self.flip_hor_cb = QtWidgets.QCheckBox()
        self.flip_ver_cb = QtWidgets.QCheckBox()

        self.ok_btn = FlatButton("Done")
        self.cancel_btn = FlatButton("Cancel")

    def _layout_widgets(self):
        self._grid_layout = QtWidgets.QGridLayout()

        self._grid_layout.addWidget(self.hor_lbl, 1, 1, 1, 1)
        self._grid_layout.addWidget(self.ver_lbl, 1, 2, 1, 1)

        self._grid_layout.addWidget(self.img_px_lbl, 2, 0, 1, 1)
        self._grid_layout.addWidget(self.img_hor_px_txt, 2, 1, 1, 1)
        self._grid_layout.addWidget(self.img_ver_px_txt, 2, 2, 1, 1)
        self._grid_layout.addWidget(self.img_px_size_lbl, 3, 0, 1, 1)
        self._grid_layout.addWidget(self.img_hor_px_size_txt, 3, 1, 1, 1)
        self._grid_layout.addWidget(self.img_ver_px_size_txt, 3, 2, 1, 1)
        self._grid_layout.addWidget(self.img_center_lbl, 4, 0, 1, 1)
        self._grid_layout.addWidget(self.img_hor_center_txt, 4, 1, 1, 1)
        self._grid_layout.addWidget(self.img_ver_center_txt, 4, 2, 1, 1)

        self._grid_layout.addWidget(self.flip_lbl, 5, 0, 1, 1)
        self._grid_layout.addWidget(self.flip_hor_cb, 5, 1, 1, 1)
        self._grid_layout.addWidget(self.flip_ver_cb, 5, 2, 1, 1)

        self._grid_layout.addWidget(self.ok_btn, 6, 1, 1, 1)
        self._grid_layout.addWidget(self.cancel_btn, 6, 2, 1, 1)

        self.setLayout(self._grid_layout)

    def _style_widgets(self):
        """
        Makes everything pretty and set double/int validators for the line edits.
        """
        self.img_hor_px_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_hor_px_txt.setMaximumWidth(60)
        self.img_ver_px_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_ver_px_txt.setMaximumWidth(60)
        self.img_hor_px_size_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_hor_px_size_txt.setMaximumWidth(60)
        self.img_ver_px_size_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_ver_px_size_txt.setMaximumWidth(60)
        self.img_hor_center_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_hor_center_txt.setMaximumWidth(60)
        self.img_ver_center_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.img_ver_center_txt.setMaximumWidth(60)

        self.img_hor_px_size_txt.setValidator(QtGui.QDoubleValidator())
        self.img_ver_px_size_txt.setValidator(QtGui.QDoubleValidator())
        self.img_hor_center_txt.setValidator(QtGui.QDoubleValidator())
        self.img_ver_center_txt.setValidator(QtGui.QDoubleValidator())
        self.img_hor_px_txt.setValidator(QtGui.QIntValidator())
        self.img_ver_px_txt.setValidator(QtGui.QIntValidator())

        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.ok_btn.setToolTip('# of files must equal # of points')

        file = open(os.path.join(style_path, "stylesheet.qss"))
        stylesheet = file.read()
        self.setStyleSheet(stylesheet)
        file.close()

    def _connect_widgets(self):
        """
        Connecting actions to slots.
        """
        self.ok_btn.clicked.connect(self.accept_bg_properties)
        self.cancel_btn.clicked.connect(self.reject_bg_properties)

    def accept_bg_properties(self):
        self.approved = True
        self.accept()

    def reject_bg_properties(self):
        self.approved = False
        self.reject()

    @property
    def hor_num_pixels(self):
        return int(str(self.img_hor_px_txt.text()))

    @property
    def ver_num_pixels(self):
        return int(str(self.img_ver_px_txt.text()))

    @property
    def hor_pixel_size(self):
        return float(str(self.img_hor_px_size_txt.text()))

    @property
    def ver_pixel_size(self):
        return float(str(self.img_ver_px_size_txt.text()))

    @property
    def hor_flip(self):
        return self.flip_hor_cb.isChecked()

    @hor_flip.setter
    def hor_flip(self, flip_value):
        self.flip_hor_cb.setChecked(flip_value)

    @property
    def ver_flip(self):
        return self.flip_ver_cb.isChecked()

    @ver_flip.setter
    def ver_flip(self, flip_value):
        self.flip_ver_cb.setChecked(flip_value)

    @property
    def hor_center(self):
        return float(str(self.img_hor_center_txt.text()))

    @hor_center.setter
    def hor_center(self, center):
        self.img_hor_center_txt.setText('{0:.4f}'.format(center))

    @property
    def ver_center(self):
        return float(str(self.img_ver_center_txt.text()))

    @ver_center.setter
    def ver_center(self, center):
        self.img_ver_center_txt.setText('{0:.4f}'.format(center))

    def exec_(self):
        """
        Overriding the dialog exec_ function to center the widget in the parent window before execution.
        """
        parent_center = self._parent.window().mapToGlobal(self._parent.window().rect().center())
        self.move(parent_center.x() - 101, parent_center.y() - 48)
        super(OpenBGImageDialog, self).exec_()


class MapErrorDialog(object):
    def __init__(self, msg):
        err_msg = QtWidgets.QMessageBox()
        err_msg.setIcon(eval(msg['icon']))
        err_msg.setText(msg['short_msg'])
        err_msg.setInformativeText(msg['informative_text'])
        err_msg.setWindowTitle(msg['window_title'])
        err_msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        err_msg.setDetailedText(msg['detailed_msg'])
        err_msg.exec_()
